r"""
Spheres smoothly embedded in Euclidean Space

Let `E^{n+1}` be a Euclidean space of dimension `n+1` and `c \in E^{n+1}`. An
`n`-sphere with radius `r` and centered at `c`, usually denoted by
`\mathbb{S}^n_r(c)`, smoothly embedded in the Euclidean space `E^{n+1}` is an
`n`-dimensional smooth manifold together with a smooth embedding

.. MATH::

    \iota \colon \mathbb{S}^n_r \to E^{n+1}

whose image consists of all points having the same Euclidean distance to the
fixed point `c`. If we choose Cartesian coordinates `(x_1, \ldots, x_{n+1})` on
`E^{n+1}` with `x(c)=0` then the above translates to

.. MATH::

    \iota(\mathbb{S}^n_r(c)) = \left\{ p \in E^{n+1} : \lVert x(p) \rVert = r \right\}.

This corresponds to the standard `n`-sphere of radius `r` centered at `c`.

AUTHORS:

- Michael Jung (2020): initial version

REFERENCES:

- \M. Berger: *Geometry I&II* [Ber1987]_, [Ber1987a]_
- \J. Lee: *Introduction to Smooth Manifolds* [Lee2013]_

EXAMPLES:

We start by defining a 2-sphere of unspecified radius `r`::

    sage: r = var('r')
    sage: S2_r = manifolds.Sphere(2, radius=r); S2_r
    2-sphere S^2_r of radius r smoothly embedded in the Euclidean space E^3

The embedding `\iota` is constructed from scratch and can be returned by the
following command::

    sage: i = S2_r.embedding(); i
    Differentiable map iota from the 2-sphere S^2_r of radius r smoothly
     embedded in the Euclidean space E^3 to the Euclidean space E^3
    sage: i.display()
    iota: S^2_r → E^3
    on A: (theta, phi) ↦ (x, y, z) = (r*cos(phi)*sin(theta),
                                         r*sin(phi)*sin(theta),
                                         r*cos(theta))

As a submanifold of a Riemannian manifold, namely the Euclidean space,
the 2-sphere admits an induced metric::

    sage: g = S2_r.induced_metric()
    sage: g.display()
    g = r^2 dtheta⊗dtheta + r^2*sin(theta)^2 dphi⊗dphi

The induced metric is also known as the *first fundamental form* (see
:meth:`~sage.manifolds.differentiable.pseudo_riemannian_submanifold.PseudoRiemannianSubmanifold.first_fundamental_form`)::

    sage: g is S2_r.first_fundamental_form()
    True

The *second fundamental form* encodes the extrinsic curvature of the
2-sphere as hypersurface of Euclidean space (see
:meth:`~sage.manifolds.differentiable.pseudo_riemannian_submanifold.PseudoRiemannianSubmanifold.second_fundamental_form`)::

    sage: K = S2_r.second_fundamental_form(); K
    Field of symmetric bilinear forms K on the 2-sphere S^2_r of radius r
     smoothly embedded in the Euclidean space E^3
    sage: K.display()
    K = r dtheta⊗dtheta + r*sin(theta)^2 dphi⊗dphi

One quantity that can be derived from the second fundamental form is the
Gaussian curvature::

    sage: K = S2_r.gauss_curvature()
    sage: K.display()
    S^2_r → ℝ
    on A: (theta, phi) ↦ r^(-2)

As we have seen, spherical coordinates are initialized by default. To
initialize stereographic coordinates retrospectively, we can use the following
command::

    sage: S2_r.stereographic_coordinates()
    Chart (S^2_r-{NP}, (y1, y2))

To get all charts corresponding to stereographic coordinates, we can use the
:meth:`~sage.manifolds.differentiable.examples.sphere.Sphere.coordinate_charts`::

    sage: stereoN, stereoS = S2_r.coordinate_charts('stereographic')
    sage: stereoN, stereoS
    (Chart (S^2_r-{NP}, (y1, y2)), Chart (S^2_r-{SP}, (yp1, yp2)))

.. SEEALSO::

    See :meth:`~sage.manifolds.differentiable.examples.sphere.Sphere.stereographic_coordinates`
    and :meth:`~sage.manifolds.differentiable.examples.sphere.Sphere.spherical_coordinates`
    for details.

.. NOTE::

    Notice that the derived quantities such as the embedding as well as the
    first and second fundamental forms must be computed from scratch again
    when new coordinates have been initialized. That makes the usage of
    previously declared objects obsolete.

Consider now a 1-sphere with barycenter `(1,0)` in Cartesian coordinates::

    sage: E2 = EuclideanSpace(2)
    sage: c = E2.point((1,0), name='c')
    sage: S1c.<chi> = E2.sphere(center=c); S1c
    1-sphere S^1(c) of radius 1 smoothly embedded in the Euclidean plane
     E^2 centered at the Point c
    sage: S1c.spherical_coordinates()
    Chart (A, (chi,))

Get stereographic coordinates::

    sage: stereoN, stereoS = S1c.coordinate_charts('stereographic')
    sage: stereoN, stereoS
    (Chart (S^1(c)-{NP}, (y1,)), Chart (S^1(c)-{SP}, (yp1,)))

The embedding takes now the following form in all coordinates::

    sage: S1c.embedding().display()
    iota: S^1(c) → E^2
    on A: chi ↦ (x, y) = (cos(chi) + 1, sin(chi))
    on S^1(c)-{NP}: y1 ↦ (x, y) = (2*y1/(y1^2 + 1) + 1, (y1^2 - 1)/(y1^2 + 1))
    on S^1(c)-{SP}: yp1 ↦ (x, y) = (2*yp1/(yp1^2 + 1) + 1, -(yp1^2 - 1)/(yp1^2 + 1))

Since the sphere is a hypersurface, we can get a normal vector field by using
``normal``::

    sage: n = S1c.normal(); n
    Vector field n along the 1-sphere S^1(c) of radius 1 smoothly embedded in
     the Euclidean plane E^2 centered at the Point c with values on the
     Euclidean plane E^2
    sage: n.display()
    n = -cos(chi) e_x - sin(chi) e_y

Notice that this is just *one* normal field with arbitrary direction,
in this particular case `n` points inwards whereas `-n` points outwards.
However, the vector field `n` is indeed non-vanishing and hence the sphere
admits an orientation (as all spheres do)::

    sage: orient = S1c.orientation(); orient
    [Coordinate frame (S^1(c)-{SP}, (∂/∂yp1)), Vector frame (S^1(c)-{NP}, (f_1))]
    sage: f = orient[1]
    sage: f[1].display()
    f_1 = -∂/∂y1

Notice that the orientation is chosen is such a way that `(\iota_*(f_1), -n)`
is oriented in the ambient Euclidean space, i.e. the last entry is the normal
vector field pointing outwards. Henceforth, the manifold admits
a volume form::

    sage: g = S1c.induced_metric()
    sage: g.display()
    g = dchi⊗dchi
    sage: eps = g.volume_form()
    sage: eps.display()
    eps_g = -dchi
"""

from sage.categories.manifolds import Manifolds
from sage.categories.metric_spaces import MetricSpaces
from sage.categories.topological_spaces import TopologicalSpaces
from sage.manifolds.differentiable.examples.euclidean import EuclideanSpace
from sage.manifolds.differentiable.pseudo_riemannian_submanifold import (
    PseudoRiemannianSubmanifold,
)
from sage.rings.real_mpfr import RR


class Sphere(PseudoRiemannianSubmanifold):
    r"""
    Sphere smoothly embedded in Euclidean Space.

    An `n`-sphere of radius `r`smoothly embedded in a Euclidean space `E^{n+1}`
    is a smooth `n`-dimensional manifold smoothly embedded into `E^{n+1}`,
    such that the embedding constitutes a standard `n`-sphere of radius `r`
    in that Euclidean space (possibly shifted by a point).

    - ``n`` -- positive integer representing dimension of the sphere
    - ``radius`` -- (default: ``1``) positive number that states the radius
      of the sphere
    - ``name`` -- (default: ``None``) string; name (symbol) given to the
      sphere; if ``None``, the name will be set according to the input
      (see convention above)
    - ``ambient_space`` -- (default: ``None``) Euclidean space in which the
      sphere should be embedded; if ``None``, a new instance of Euclidean
      space is created
    - ``center`` -- (default: ``None``) the barycenter of the sphere as point of
      the ambient Euclidean space; if ``None`` the barycenter is set to the
      origin of the ambient space's standard Cartesian coordinates
    - ``latex_name`` -- (default: ``None``) string; LaTeX symbol to
      denote the space; if ``None``, it will be set according to the input
      (see convention above)
    - ``coordinates`` -- (default: ``'spherical'``) string describing the
      type of coordinates to be initialized at the sphere's creation; allowed
      values are

          - ``'spherical'`` spherical coordinates (see
            :meth:`~sage.manifolds.differentiable.examples.sphere.Sphere.spherical_coordinates`))
          - ``'stereographic'`` stereographic coordinates given by the
            stereographic projection (see
            :meth:`~sage.manifolds.differentiable.examples.sphere.Sphere.stereographic_coordinates`)

    - ``names`` -- (default: ``None``) must be a tuple containing
      the coordinate symbols (this guarantees the shortcut operator
      ``<,>`` to function); if ``None``, the usual conventions are used (see
      examples below for details)
    - ``unique_tag`` -- (default: ``None``) tag used to force the construction
      of a new object when all the other arguments have been used previously
      (without ``unique_tag``, the
      :class:`~sage.structure.unique_representation.UniqueRepresentation`
      behavior inherited from
      :class:`~sage.manifolds.differentiable.pseudo_riemannian.PseudoRiemannianManifold`
      would return the previously constructed object corresponding to these
      arguments)

    EXAMPLES:

    A 2-sphere embedded in Euclidean space::

        sage: S2 = manifolds.Sphere(2); S2
        2-sphere S^2 of radius 1 smoothly embedded in the Euclidean space E^3
        sage: latex(S2)
        \mathbb{S}^{2}

    The ambient Euclidean space is constructed incidentally::

        sage: S2.ambient()
        Euclidean space E^3

    Another call creates another sphere and hence another Euclidean space::

        sage: S2 is manifolds.Sphere(2)
        False
        sage: S2.ambient() is manifolds.Sphere(2).ambient()
        False

    By default, the barycenter is set to the coordinate origin of the
    standard Cartesian coordinates in the ambient Euclidean space::

        sage: c = S2.center(); c
        Point on the Euclidean space E^3
        sage: c.coord()
        (0, 0, 0)

    Each `n`-sphere is a compact manifold and a complete metric space::

        sage: S2.category()
        Join of Category of compact topological spaces and Category of smooth
         manifolds over Real Field with 53 bits of precision and Category of
         connected manifolds over Real Field with 53 bits of precision and
         Category of complete metric spaces

    If not stated otherwise, each `n`-sphere is automatically endowed with
    spherical coordinates::

        sage: S2.atlas()
        [Chart (A, (theta, phi))]
        sage: S2.default_chart()
        Chart (A, (theta, phi))
        sage: spher = S2.spherical_coordinates()
        sage: spher is S2.default_chart()
        True

    Notice that the spherical coordinates do not cover the whole sphere. To
    cover the entire sphere with charts, use stereographic coordinates instead::

        sage: stereoN, stereoS = S2.coordinate_charts('stereographic')
        sage: stereoN, stereoS
        (Chart (S^2-{NP}, (y1, y2)), Chart (S^2-{SP}, (yp1, yp2)))
        sage: list(S2.open_covers())
        [Set {S^2} of open subsets of the 2-sphere S^2 of radius 1 smoothly embedded in the Euclidean space E^3,
         Set {S^2-{NP}, S^2-{SP}} of open subsets of the 2-sphere S^2 of radius 1 smoothly embedded in the Euclidean space E^3]

    .. NOTE::

        Keep in mind that the initialization process of stereographic
        coordinates and their transition maps is computational complex in
        higher dimensions. Henceforth, high computation times are expected with
        increasing dimension.
    """
    @staticmethod
    def __classcall_private__(cls, n=None, radius=1, ambient_space=None,
                              center=None, name=None, latex_name=None,
                              coordinates='spherical', names=None,
                              unique_tag=None):
        r"""
        Determine the correct class to return based upon the input.

        TESTS:

        Each call gives a new instance::

            sage: S2 = manifolds.Sphere(2)
            sage: S2 is manifolds.Sphere(2)
            False

        The dimension can be determined using the ``<...>`` operator::

            sage: S.<x,y> = manifolds.Sphere(coordinates='stereographic'); S
            2-sphere S^2 of radius 1 smoothly embedded in the Euclidean space E^3
            sage: S._first_ngens(2)
            (x, y)
        """
        if n is None:
            if names is None:
                raise ValueError("either n or names must be specified")
            n = len(names)

        # Technical bit for UniqueRepresentation
        from time import time

        from sage.misc.prandom import getrandbits
        if unique_tag is None:
            unique_tag = getrandbits(128) * time()

        return super().__classcall__(cls, n, radius=radius,
                                     ambient_space=ambient_space,
                                     center=center,
                                     name=name, latex_name=latex_name,
                                     coordinates=coordinates, names=names,
                                     unique_tag=unique_tag)

    def __init__(self, n, radius=1, ambient_space=None, center=None, name=None,
                 latex_name=None, coordinates='spherical', names=None,
                 category=None, init_coord_methods=None, unique_tag=None):
        r"""
        Construct sphere smoothly embedded in Euclidean space.

        TESTS::

            sage: S2 = manifolds.Sphere(2); S2
            2-sphere S^2 of radius 1 smoothly embedded in the Euclidean space E^3
            sage: S2.metric()
            Riemannian metric g on the 2-sphere S^2 of radius 1 smoothly
             embedded in the Euclidean space E^3
            sage: TestSuite(S2).run()
        """
        # radius
        if radius <= 0:
            raise ValueError('radius must be greater than zero')
        # ambient space
        if ambient_space is None:
            ambient_space = EuclideanSpace(n+1)
        elif not isinstance(ambient_space, EuclideanSpace):
            raise TypeError("the argument 'ambient_space' must be a Euclidean space")
        elif ambient_space._dim != n+1:
            raise ValueError("Euclidean space must have dimension {}".format(n+1))
        if center is None:
            cart = ambient_space.cartesian_coordinates()
            c_coords = [0]*(n+1)
            center = ambient_space.point(c_coords, chart=cart)
        elif center not in ambient_space:
            raise ValueError('{} must be an element of {}'.format(center, ambient_space))
        if name is None:
            name = 'S^{}'.format(n)
            if radius != 1:
                name += r'_{}'.format(radius)
            if center._name:
                name += r'({})'.format(center._name)
            if latex_name is None:
                latex_name = r'\mathbb{S}^{' + str(n) + r'}'
                if radius != 1:
                    latex_name += r'_{{{}}}'.format(radius)
                if center._latex_name:
                    latex_name += r'({})'.format(center._latex_name)
        if category is None:
            category = Manifolds(RR).Smooth() & MetricSpaces().Complete() & \
                       TopologicalSpaces().Compact().Connected()
        # initialize
        PseudoRiemannianSubmanifold.__init__(self, n, name,
                                             ambient=ambient_space,
                                             signature=n, latex_name=latex_name,
                                             metric_name='g', start_index=1,
                                             category=category)
        # set attributes
        self._radius = radius
        self._center = center
        self._coordinates = {}  # established coordinates; values are lists
        self._init_coordinates = {'spherical': self._init_spherical,
                                  'stereographic': self._init_stereographic}
                                 # predefined coordinates
        if init_coord_methods:
            self._init_coordinates.update(init_coord_methods)
        if coordinates not in self._init_coordinates:
            raise ValueError('{} coordinates not available'.format(coordinates))
        # up here, the actual initialization begins:
        self._init_chart_domains()
        self._init_embedding()
        self._init_coordinates[coordinates](names)

    def _init_embedding(self):
        r"""
        Initialize the embedding into Euclidean space.

        TESTS::

            sage: S2 = manifolds.Sphere(2)
            sage: i = S2.embedding(); i
            Differentiable map iota from the 2-sphere S^2 of radius 1 smoothly
             embedded in the Euclidean space E^3 to the Euclidean space E^3
            sage: i.display()
            iota: S^2 → E^3
             on A: (theta, phi) ↦ (x, y, z) = (cos(phi)*sin(theta),
             sin(phi)*sin(theta), cos(theta))
        """
        name = 'iota'
        latex_name = r'\iota'
        iota = self.diff_map(self._ambient, name=name, latex_name=latex_name)
        self.set_embedding(iota)

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        TESTS::

            sage: S2_3 = manifolds.Sphere(2, radius=3)
            sage: S2_3._repr_()
            '2-sphere S^2_3 of radius 3 smoothly embedded in the Euclidean space E^3'
            sage: S2_3  # indirect doctest
            2-sphere S^2_3 of radius 3 smoothly embedded in the Euclidean space E^3
        """
        s = "{}-sphere {} of radius {} smoothly embedded in " \
            "the {}".format(self._dim, self._name, self._radius, self._ambient)
        if self._center._name:
            s += ' centered at the Point {}'.format(self._center._name)
        return s

    def coordinate_charts(self, coord_name, names=None):
        r"""
        Return a list of all charts belonging to the coordinates ``coord_name``.

        INPUT:

        - ``coord_name`` -- string describing the type of coordinates
        - ``names`` -- (default: ``None``) must be a tuple containing
          the coordinate symbols for the first chart in the list; if
          ``None``, the standard convention is used

        EXAMPLES:

        Spherical coordinates on `S^1`::

            sage: S1 = manifolds.Sphere(1)
            sage: S1.coordinate_charts('spherical')
            [Chart (A, (phi,))]

        Stereographic coordinates on `S^1`::

            sage: stereo_charts = S1.coordinate_charts('stereographic', names=['a'])
            sage: stereo_charts
            [Chart (S^1-{NP}, (a,)), Chart (S^1-{SP}, (ap,))]
        """
        if coord_name not in self._coordinates:
            if coord_name not in self._init_coordinates:
                raise ValueError('{} coordinates not available'.format(coord_name))
            self._init_coordinates[coord_name](names)
        return list(self._coordinates[coord_name])

    def _first_ngens(self, n):
        r"""
        Return the list of coordinates of the default chart.

        This is useful only for the use of Sage preparser::

            sage: preparse("S3.<x,y,z> = manifolds.Sphere(coordinates='stereographic')")
            "S3 = manifolds.Sphere(coordinates='stereographic', names=('x', 'y', 'z',));
             (x, y, z,) = S3._first_ngens(3)"

        TESTS::

            sage: S2 = manifolds.Sphere(2)
            sage: S2._first_ngens(2)
            (theta, phi)
            sage: S2.<u,v> = manifolds.Sphere(2)
            sage: S2._first_ngens(2)
            (u, v)
        """
        return self._def_chart[:]

    def _init_chart_domains(self):
        r"""
        Construct the chart domains on ``self``.

        TESTS::

            sage: S2 = manifolds.Sphere(2)
            sage: list(S2.open_covers())
            [Set {S^2} of open subsets of the 2-sphere S^2 of radius 1 smoothly embedded in the Euclidean space E^3,
             Set {S^2-{NP}, S^2-{SP}} of open subsets of the 2-sphere S^2 of radius 1 smoothly embedded in the Euclidean space E^3]
            sage: frozenset(S2.subsets())  # random
            frozenset({Euclidean 2-sphere S^2 of radius 1,
             Open subset A of the Euclidean 2-sphere S^2 of radius 1,
             Open subset S^2-{NP,SP} of the Euclidean 2-sphere S^2 of radius 1,
             Open subset S^2-{NP} of the Euclidean 2-sphere S^2 of radius 1,
             Open subset S^2-{SP} of the Euclidean 2-sphere S^2 of radius 1})
        """
        # without north pole:
        name = self._name + '-{NP}'
        latex_name = self._latex_name + r'\setminus\{\mathrm{NP}\}'
        self._stereoN_dom = self.open_subset(name, latex_name)
        # without south pole:
        name = self._name + '-{SP}'
        latex_name = self._latex_name + r'\setminus\{\mathrm{SP}\}'
        self._stereoS_dom = self.open_subset(name, latex_name)
        # intersection:
        int = self._stereoN_dom.intersection(self._stereoS_dom)
        int._name = self._name + '-{NP,SP}'
        int._latex_name = self._latex_name + \
                          r'\setminus\{\mathrm{NP}, \mathrm{SP}\}'
        # without half circle:
        self._spher_dom = int.open_subset('A')
        # declare union:
        self.declare_union(self._stereoN_dom, self._stereoS_dom)

    def _shift_coords(self, coordfunc, s='+'):
        r"""
        Shift the coordinates ``coordfunc`` given in Cartesian coordinates to
        the barycenter.

        INPUT:

        - ``coordfunc`` -- coordinate expression given in the standard
          Cartesian coordinates of the ambient Euclidean space
        - ``s`` -- either ``'+'`` or ``'-'`` depending on whether ``coordfunc``
          should be shifted from the coordinate origin to the center or from
          the center to the coordinate origin

        TESTS::

            sage: E3 = EuclideanSpace(3)
            sage: c = E3.point((1,2,3), name='c')
            sage: S2c = manifolds.Sphere(2, ambient_space=E3, center=c)
            sage: spher = S2c.spherical_coordinates()
            sage: cart = E3.cartesian_coordinates()
            sage: coordfunc = S2c.embedding().expr(spher, cart); coordfunc
            (cos(phi)*sin(theta) + 1, sin(phi)*sin(theta) + 2, cos(theta) + 3)
            sage: S2c._shift_coords(coordfunc, s='-')
            (cos(phi)*sin(theta), sin(phi)*sin(theta), cos(theta))
        """
        cart = self._ambient.cartesian_coordinates()
        c_coords = cart(self._center)
        if s == '+':
            res = [x + c for x, c in zip(coordfunc, c_coords)]
        elif s == '-':
            res = [x - c for x, c in zip(coordfunc, c_coords)]
        return tuple(res)

    def _init_spherical(self, names=None):
        r"""
        Construct the chart of spherical coordinates.

        TESTS:

        Spherical coordinates on a 2-sphere::

            sage: S2 = manifolds.Sphere(2)
            sage: S2.spherical_coordinates()
            Chart (A, (theta, phi))

        Spherical coordinates on a 1-sphere::

            sage: S1 = manifolds.Sphere(1, coordinates='stereographic')
            sage: spher = S1.spherical_coordinates(); spher  # create coords
            Chart (A, (phi,))
            sage: S1.atlas()
            [Chart (S^1-{NP}, (y1,)),
             Chart (S^1-{SP}, (yp1,)),
             Chart (S^1-{NP,SP}, (y1,)),
             Chart (S^1-{NP,SP}, (yp1,)),
             Chart (A, (phi,)),
             Chart (A, (y1,)),
             Chart (A, (yp1,))]
        """
        # speed-up via simplification method...
        self.set_simplify_function(lambda expr: expr.simplify_trig())

        # get domain...
        A = self._spher_dom

        # initialize coordinates...
        n = self._dim
        if names:
            # add interval:
            names = tuple([x + ':(0,pi)' for x in names[:-1]] +
                          [names[-1] + ':(-pi,pi):periodic'])
        else:
            if n == 1:
                names = ('phi:(-pi,pi):periodic',)
            elif n == 2:
                names = ('theta:(0,pi)', 'phi:(-pi,pi):periodic')
            elif n == 3:
                names = ('chi:(0,pi)', 'theta:(0,pi)', 'phi:(-pi,pi):periodic')
            else:
                names = tuple(["phi_{}:(0,pi)".format(i) for i in range(1,n)] +
                              ["phi_{}:(-pi,pi):periodic".format(n)])
        spher = A.chart(names=names)
        coord = spher[:]

        # make spherical chart and frame the default ones on their domain:
        A.set_default_chart(spher)
        A.set_default_frame(spher.frame())

        # manage embedding...
        from sage.functions.trig import cos, sin
        from sage.misc.misc_c import prod

        R = self._radius

        coordfunc = [R*cos(coord[n-1])*prod(sin(coord[i]) for i in range(n-1))]
        coordfunc += [R*prod(sin(coord[i]) for i in range(n))]
        for k in reversed(range(n-1)):
            c = R*cos(coord[k])*prod(sin(coord[i]) for i in range(k))
            coordfunc.append(c)
        cart = self._ambient.cartesian_coordinates()
        # shift coordinates to barycenter:
        coordfunc = self._shift_coords(coordfunc, s='+')
        # add expression to embedding:
        self._immersion.add_expr(spher, cart, coordfunc)
        self.clear_cache()  # clear cache of extrinsic information

        # finish process...
        self._coordinates['spherical'] = [spher]

        # adapt other coordinates...
        if 'stereographic' in self._coordinates:
            self._transition_spher_stereo()

        # reset simplification method...
        self.set_simplify_function('default')

    def stereographic_coordinates(self, pole='north', names=None):
        r"""
        Return stereographic coordinates given by the stereographic
        projection of ``self`` w.r.t. to a given pole.

        INPUT:

        - ``pole`` -- (default: ``'north'``) the pole determining the
          stereographic projection; possible options are ``'north'`` and
          ``'south'``
        - ``names`` -- (default: ``None``) must be a tuple containing
          the coordinate symbols (this guarantees the usage of the shortcut
          operator ``<,>``)

        OUTPUT:

        - the chart of stereographic coordinates w.r.t. to the given pole,
          as an instance of
          :class:`~sage.manifolds.differentiable.chart.RealDiffChart`

        Let `\mathbb{S}^n_r(c)` be an `n`-sphere of radius `r` smoothly
        embedded in the Euclidean space `E^{n+1}` centered at `c \in E^{n+1}`.
        We denote the north pole of `\mathbb{S}^n_r(c)` by `\mathrm{NP}` and the
        south pole by `\mathrm{SP}`. These poles are uniquely determined by
        the requirement

        .. MATH::

            x(\iota(\mathrm{NP})) &= (0, \ldots, 0, r) + x(c), \\
            x(\iota(\mathrm{SP})) &= (0, \ldots, 0, -r) + x(c).

        The coordinates `(y_1, \ldots, y_n)` (`(y'_1, \ldots, y'_n)`
        respectively) define *stereographic coordinates* on `\mathbb{S}^n_r(c)`
        for the Cartesian coordinates `(x_1, \ldots, x_{n+1})` on `E^{n+1}`
        if they arise from the stereographic projection from
        `\iota(\mathrm{NP})` (`\iota(\mathrm{SP})`) to the hypersurface
        `x_n = x_n(c)`. In concrete formulas, this means:

        .. MATH::

            \left. x \circ \iota \right|_{\mathbb{S}^n_r(c) \setminus \{
            \mathrm{NP}\}}
            &= \left( \frac{2y_1r^2}{r^2+\sum^n_{i=1} y^2_i},
            \ldots, \frac{2y_nr^2}{r^2+\sum^n_{i=1} y^2_i},
            \frac{r\sum^n_{i=1} y^2_i-r^3}{r^2+\sum^n_{i=1} y^2_i} \right) +
            x(c), \\
            \left. x \circ \iota \right|_{\mathbb{S}^n_r(c) \setminus \{
            \mathrm{SP}\}} &= \left( \frac{2y'_1r^2}{r^2+\sum^n_{i=1} y'^2_i},
            \ldots, \frac{2y'_nr^2}{r^2+\sum^n_{i=1} y'^2_i},
            \frac{r^3 - r\sum^n_{i=1} y'^2_i}{r^2+\sum^n_{i=1} y'^2_i} \right) +
            x(c).

        EXAMPLES:

        Initialize a 1-sphere centered at `(1,0)` in the Euclidean plane
        using the shortcut operator::

            sage: E2 = EuclideanSpace(2)
            sage: c = E2.point((1,0), name='c')
            sage: S1.<a> = E2.sphere(center=c, coordinates='stereographic'); S1
            1-sphere S^1(c) of radius 1 smoothly embedded in the Euclidean plane
             E^2 centered at the Point c

        By default, the shortcut variables belong to the stereographic
        projection from the north pole::

            sage: S1.coordinate_charts('stereographic')
            [Chart (S^1(c)-{NP}, (a,)), Chart (S^1(c)-{SP}, (ap,))]
            sage: S1.embedding().display()
            iota: S^1(c) → E^2
            on S^1(c)-{NP}: a ↦ (x, y) = (2*a/(a^2 + 1) + 1, (a^2 - 1)/(a^2 + 1))
            on S^1(c)-{SP}: ap ↦ (x, y) = (2*ap/(ap^2 + 1) + 1, -(ap^2 - 1)/(ap^2 + 1))

        Initialize a 2-sphere from scratch::

            sage: S2 = manifolds.Sphere(2)
            sage: S2.atlas()
            [Chart (A, (theta, phi))]

        In the previous block, the stereographic coordinates have not been
        initialized. This happens subsequently with the invocation of
        ``stereographic_coordinates``::

            sage: stereoS.<u,v> = S2.stereographic_coordinates(pole='south')
            sage: S2.coordinate_charts('stereographic')
            [Chart (S^2-{NP}, (up, vp)), Chart (S^2-{SP}, (u, v))]

        If not specified by the user, the default coordinate names are given by
        `(y_1, \ldots, y_n)` and `(y'_1, \ldots, y'_n)` respectively::

            sage: S3 = manifolds.Sphere(3, coordinates='stereographic')
            sage: S3.stereographic_coordinates(pole='north')
            Chart (S^3-{NP}, (y1, y2, y3))
            sage: S3.stereographic_coordinates(pole='south')
            Chart (S^3-{SP}, (yp1, yp2, yp3))
        """
        coordinates = 'stereographic'
        if coordinates not in self._coordinates:
            self._init_coordinates[coordinates](names, default_pole=pole)
        if pole == 'north':
            return self._coordinates[coordinates][0]
        if pole == 'south':
            return self._coordinates[coordinates][1]
        raise ValueError("pole must be 'north' or 'south'")

    def spherical_coordinates(self, names=None):
        r"""
        Return the spherical coordinates of ``self``.

        INPUT:

        - ``names`` -- (default: ``None``) must be a tuple containing
          the coordinate symbols (this guarantees the usage of the shortcut
          operator ``<,>``)

        OUTPUT:

        - the chart of spherical coordinates, as an instance of
          :class:`~sage.manifolds.differentiable.chart.RealDiffChart`

        Let `\mathbb{S}^n_r(c)` be an `n`-sphere of radius `r` smoothly
        embedded in the Euclidean space `E^{n+1}` centered at `c \in E^{n+1}`.
        We say that `(\varphi_1, \ldots, \varphi_n)` define *spherical
        coordinates* on the open subset `A \subset \mathbb{S}^n_r(c)` for the
        Cartesian coordinates `(x_1, \ldots, x_{n+1})` on `E^{n+1}` (not
        necessarily centered at `c`) if

        .. MATH::

            \begin{aligned}
                \left. x_1 \circ \iota \right|_{A} &= r \cos(\varphi_n)\sin(
                \varphi_{n-1}) \cdots
                \sin(\varphi_1)
                + x_1(c), \\
                \left. x_1 \circ \iota \right|_{A} &= r \sin(\varphi_n)\sin(
                \varphi_{n-1}) \cdots
                \sin(\varphi_1)
                + x_1(c), \\
                \left. x_2 \circ \iota \right|_{A} &= r \cos(\varphi_{
                n-1})\sin(\varphi_{n-2}) \cdots
                \sin(\varphi_1)
                + x_2(c), \\
                \left. x_3 \circ \iota \right|_{A} &= r \cos(\varphi_{
                n-2})\sin(\varphi_{n-3}) \cdots
                \sin(\varphi_1)
                + x_3(c), \\
                \vdots & \\
                \left. x_{n+1} \circ \iota \right|_{A} &= r \cos(\varphi_1) +
                x_{n+1}(c),
            \end{aligned}

        where `\varphi_i` has range `(0, \pi)` for `i=1, \ldots, n-1` and
        `\varphi_n` lies in `(-\pi, \pi)`. Notice that the above expressions
        together with the ranges of the `\varphi_i` fully determine the open
        set `A`.

        .. NOTE::

            Notice that our convention slightly differs from the one given on
            the :wikipedia:`N-sphere#Spherical_coordinates`. The definition
            above ensures that the conventions for the most common cases
            `n=1` and `n=2` are maintained.

        EXAMPLES:

        The spherical coordinates on a 2-sphere follow the common conventions::

            sage: S2 = manifolds.Sphere(2)
            sage: spher = S2.spherical_coordinates(); spher
            Chart (A, (theta, phi))

        The coordinate range of spherical coordinates::

            sage: spher.coord_range()
            theta: (0, pi); phi: [-pi, pi] (periodic)

        Spherical coordinates do not cover the 2-sphere entirely::

            sage: A = spher.domain(); A
            Open subset A of the 2-sphere S^2 of radius 1 smoothly embedded in
             the Euclidean space E^3

        The embedding of a 2-sphere in Euclidean space via spherical
        coordinates::

            sage: S2.embedding().display()
            iota: S^2 → E^3
             on A: (theta, phi) ↦ (x, y, z) =
                                     (cos(phi)*sin(theta),
                                      sin(phi)*sin(theta),
                                      cos(theta))

        Now, consider spherical coordinates on a 3-sphere::

            sage: S3 = manifolds.Sphere(3)
            sage: spher = S3.spherical_coordinates(); spher
            Chart (A, (chi, theta, phi))
            sage: S3.embedding().display()
            iota: S^3 → E^4
            on A: (chi, theta, phi) ↦ (x1, x2, x3, x4) =
                                         (cos(phi)*sin(chi)*sin(theta),
                                          sin(chi)*sin(phi)*sin(theta),
                                          cos(theta)*sin(chi),
                                          cos(chi))

        By convention, the last coordinate is periodic::

            sage: spher.coord_range()
            chi: (0, pi); theta: (0, pi); phi: [-pi, pi] (periodic)
        """
        coordinates = 'spherical'
        if coordinates not in self._coordinates:
            self._init_coordinates[coordinates](names)
        return self._coordinates[coordinates][0]

    def _init_stereographic(self, names, default_pole='north'):
        r"""
        Construct the charts of stereographic coordinates.

        TESTS:

        Stereographic coordinates on the 2-sphere::

            sage: S2.<x,y> = manifolds.Sphere(2, coordinates='stereographic')
            sage: S2.atlas()
            [Chart (S^2-{NP}, (x, y)),
             Chart (S^2-{SP}, (xp, yp)),
             Chart (S^2-{NP,SP}, (x, y)),
             Chart (S^2-{NP,SP}, (xp, yp))]

        Stereographic coordinates on the 1-sphere::

            sage: S1 = manifolds.Sphere(1)
            sage: stereoS.<x> = S1.stereographic_coordinates(pole='south')
            sage: S1.atlas()
            [Chart (A, (phi,)),
             Chart (S^1-{NP}, (xp,)),
             Chart (S^1-{SP}, (x,)),
             Chart (S^1-{NP,SP}, (xp,)),
             Chart (S^1-{NP,SP}, (x,)),
             Chart (A, (xp,)),
             Chart (A, (x,))]

        The stereographic chart is the default one on its domain::

            sage: V = stereoS.domain()
            sage: V.default_chart()
            Chart (S^1-{SP}, (x,))

        Accordingly, we have::

            sage: S1.metric().restrict(V).display()
            g = 4/(x^4 + 2*x^2 + 1) dx⊗dx

        while the spherical chart is still the default one on ``S1``::

            sage: S1.metric().display()
            g = dphi⊗dphi
        """
        # speed-up via simplification method...
        self.set_simplify_function(lambda expr: expr.simplify_rational())
        # TODO: More speed-up?

        # get domains...
        U = self._stereoN_dom
        V = self._stereoS_dom

        # initialize coordinates...
        symbols_N = ''
        symbols_S = ''
        if names:
            for x in names:
                if default_pole == 'north':
                    symbols_N += x + ' '
                    symbols_S += "{}p".format(x) + ":{}' ".format(x)
                elif default_pole == 'south':
                    symbols_S += x + ' '
                    symbols_N += "{}p".format(x) + ":{}' ".format(x)
        else:
            for i in self.irange():
                symbols_N += "y{}".format(i) + r":y_{" + str(i) + r"} "
                symbols_S += "yp{}".format(i) + r":y'_{" + str(i) + r"} "
        symbols_N = symbols_N[:-1]
        symbols_S = symbols_S[:-1]
        stereoN = U.chart(coordinates=symbols_N)
        stereoS = V.chart(coordinates=symbols_S)
        coordN = stereoN[:]
        coordS = stereoS[:]

        # make stereographic charts and frames the default ones on their
        # respective domains:
        U.set_default_chart(stereoN)
        V.set_default_chart(stereoS)
        U.set_default_frame(stereoN.frame())
        V.set_default_frame(stereoS.frame())

        # predefine variables...
        r2_N = sum(y ** 2 for y in coordN)
        r2_S = sum(yp ** 2 for yp in coordS)
        R = self._radius
        R2 = R**2

        # define transition map...
        coordN_to_S = tuple(R*y/r2_N for y in coordN)
        coordS_to_N = tuple(R*yp/r2_S for yp in coordS)
        stereoN_to_S = stereoN.transition_map(stereoS, coordN_to_S,
                                              restrictions1=r2_N != 0,
                                              restrictions2=r2_S != 0)
        stereoN_to_S.set_inverse(*coordS_to_N, check=False)

        # manage embedding...
        coordfuncN = [2*y*R2 / (R2+r2_N) for y in coordN]
        coordfuncN += [(R*r2_N-R*R2)/(R2+r2_N)]
        coordfuncS = [2*yp*R2 / (R2+r2_S) for yp in coordS]
        coordfuncS += [(R*R2-R*r2_S)/(R2+r2_S)]
        cart = self._ambient.cartesian_coordinates()
        # shift coordinates to barycenter:
        coordfuncN = self._shift_coords(coordfuncN, s='+')
        coordfuncS = self._shift_coords(coordfuncS, s='+')
        # add expressions to embedding:
        self._immersion.add_expr(stereoN, cart, coordfuncN)
        self._immersion.add_expr(stereoS, cart, coordfuncS)
        self.clear_cache()  # clear cache of extrinsic information

        # define orientation...
        eN = stereoN.frame()
        eS = stereoS.frame()  # oriented w.r.t. embedding
        frame_comp = list(eN[:])
        frame_comp[0] = -frame_comp[0]  # reverse orientation
        f = U.vector_frame('f', frame_comp)
        self.set_orientation([eS, f])

        # finish process...
        self._coordinates['stereographic'] = [stereoN, stereoS]

        # adapt other coordinates...
        if 'spherical' in self._coordinates:
            self._transition_spher_stereo()

        # reset simplification method...
        self.set_simplify_function('default')

    def _transition_spher_stereo(self):
        r"""
        Initialize the transition map between spherical and stereographic
        coordinates.

        TESTS::

            sage: S1 = manifolds.Sphere(1)
            sage: spher = S1.spherical_coordinates(); spher
            Chart (A, (phi,))
            sage: A = spher.domain()
            sage: stereoN = S1.stereographic_coordinates(pole='north'); stereoN
            Chart (S^1-{NP}, (y1,))
            sage: S1.coord_change(spher, stereoN.restrict(A))
            Change of coordinates from Chart (A, (phi,)) to Chart (A, (y1,))
        """
        # speed-up via simplification method...
        self.set_simplify_function(lambda expr: expr.simplify())

        # configure preexisting charts...
        W = self._stereoN_dom.intersection(self._stereoS_dom)
        A = self._spher_dom
        stereoN, stereoS = self._coordinates['stereographic'][:]
        coordN = stereoN[:]
        coordS = stereoS[:]
        rstN = (coordN[0] != 0,)
        rstS = (coordS[0] != 0,)
        if self._dim > 1:
            rstN += (coordN[0] > 0,)
            rstS += (coordS[0] > 0,)
        stereoN_A = stereoN.restrict(A, rstN)
        stereoS_A = stereoS.restrict(A, rstS)
        self._coord_changes[(stereoN.restrict(W),
                             stereoS.restrict(W))].restrict(A)
        self._coord_changes[(stereoS.restrict(W),
                             stereoN.restrict(W))].restrict(A)
        spher = self._coordinates['spherical'][0]

        R = self._radius
        n = self._dim

        # transition: spher to stereoN...
        imm = self.embedding()
        cart = self._ambient.cartesian_coordinates()
        # get ambient coordinates and shift to coordinate origin:
        x = self._shift_coords(imm.expr(spher, cart), s='-')
        coordfunc = [(R*x[i])/(R-x[-1]) for i in range(n)]
        # define transition map:
        spher_to_stereoN = spher.transition_map(stereoN_A, coordfunc)

        # transition: stereoN to spher...
        from sage.functions.trig import acos, atan2
        from sage.misc.functional import sqrt
        # get ambient coordinates and shift to coordinate origin:
        x = self._shift_coords(imm.expr(stereoN, cart), s='-')
        coordfunc = [atan2(x[1],x[0])]
        for k in range(2, n+1):
            c = acos(x[k]/sqrt(sum(x[i]**2 for i in range(k+1))))
            coordfunc.append(c)
        coordfunc = reversed(coordfunc)
        spher_to_stereoN.set_inverse(*coordfunc, check=False)

        # transition spher <-> stereoS...
        stereoN_to_S_A = self.coord_change(stereoN_A, stereoS_A)
        stereoN_to_S_A * spher_to_stereoN  # generates spher_to_stereoS
        stereoS_to_N_A = self.coord_change(stereoS_A, stereoN_A)
        spher_to_stereoN.inverse() * stereoS_to_N_A  # generates stereoS_to_spher

    def dist(self, p, q):
        r"""
        Return the great circle distance between the points ``p`` and ``q`` on
        ``self``.

        INPUT:

        - ``p`` -- an element of ``self``
        - ``q`` -- an element of ``self``

        OUTPUT:

        - the great circle distance `d(p, q)` on ``self``

        The great circle distance `d(p, q)` of the points
        `p, q \in \mathbb{S}^n_r(c)` is the length of the shortest great circle
        segment on `\mathbb{S}^n_r(c)` that joins `p` and `q`. If we choose
        Cartesian coordinates `(x_1, \ldots, x_{n+1})` of the ambient Euclidean
        space such that the center lies in the coordinate origin, i.e.
        `x(c)=0`, the great circle distance can be expressed in terms of the
        following formula:

        .. MATH::

            d(p,q) = r \, \arccos\left(\frac{x(\iota(p)) \cdot
                x(\iota(q))}{r^2}\right).

        EXAMPLES:

        Define a 2-sphere with unspecified radius::

            sage: r = var('r')
            sage: S2_r = manifolds.Sphere(2, radius=r); S2_r
            2-sphere S^2_r of radius r smoothly embedded in the Euclidean space E^3

        Given two antipodal points in spherical coordinates::

            sage: p = S2_r.point((pi/2, pi/2), name='p'); p
            Point p on the 2-sphere S^2_r of radius r smoothly embedded in the
             Euclidean space E^3
            sage: q = S2_r.point((pi/2, -pi/2), name='q'); q
            Point q on the 2-sphere S^2_r of radius r smoothly embedded in the
             Euclidean space E^3

        The distance is determined as the length of the half great circle::

            sage: S2_r.dist(p, q)
            pi*r
        """
        from sage.functions.trig import acos
        # get Euclidean points:
        x = self._immersion(p)
        y = self._immersion(q)
        cart = self._ambient.cartesian_coordinates()
        # get ambient coordinates and shift to coordinate origin:
        x_coord = self._shift_coords(x.coord(chart=cart), s='-')
        y_coord = self._shift_coords(y.coord(chart=cart), s='-')

        n = self._dim + 1
        r = self._radius
        inv_angle = sum(x_coord[i]*y_coord[i] for i in range(n)) / r**2
        return (r * acos(inv_angle)).simplify()

    def radius(self):
        r"""
        Return the radius of ``self``.

        EXAMPLES:

        3-sphere with radius 3::

            sage: S3_2 = manifolds.Sphere(3, radius=2); S3_2
            3-sphere S^3_2 of radius 2 smoothly embedded in the 4-dimensional
             Euclidean space E^4
            sage: S3_2.radius()
            2

        2-sphere with unspecified radius::

            sage: r = var('r')
            sage: S2_r = manifolds.Sphere(3, radius=r); S2_r
            3-sphere S^3_r of radius r smoothly embedded in the 4-dimensional
             Euclidean space E^4
            sage: S2_r.radius()
            r
        """
        return self._radius

    def minimal_triangulation(self):
        r"""
        Return the minimal triangulation of ``self`` as a simplicial complex.

        EXAMPLES:

        Minimal triangulation of the 2-sphere::

            sage: S2 = manifolds.Sphere(2)
            sage: S = S2.minimal_triangulation(); S
            Minimal triangulation of the 2-sphere

        The Euler characteristic of a 2-sphere::

            sage: S.euler_characteristic()
            2
        """
        from sage.topology.simplicial_complex_examples import Sphere as SymplicialSphere
        return SymplicialSphere(self._dim)

    def center(self):
        r"""
        Return the barycenter of ``self`` in the ambient Euclidean space.

        EXAMPLES:

        2-sphere embedded in Euclidean space centered at `(1,2,3)` in
        Cartesian coordinates::

            sage: E3 = EuclideanSpace(3)
            sage: c = E3.point((1,2,3), name='c')
            sage: S2c = manifolds.Sphere(2, ambient_space=E3, center=c); S2c
            2-sphere S^2(c) of radius 1 smoothly embedded in the Euclidean space
             E^3 centered at the Point c
            sage: S2c.center()
            Point c on the Euclidean space E^3

        We can see that the embedding is shifted accordingly::

            sage: S2c.embedding().display()
            iota: S^2(c) → E^3
            on A: (theta, phi) ↦ (x, y, z) = (cos(phi)*sin(theta) + 1,
                                                 sin(phi)*sin(theta) + 2,
                                                 cos(theta) + 3)
        """
        return self._center



# distutils: libraries = homfly gc
r"""
Cython wrapper for libhomfly library

This is used to call the libhomfly library directly from python. Knots
and links are passed following the convention in libhomfly. It is basically
the oriented Gauss code, represented as a string of integers separated
by spaces as follows:

- how many strings,

  - for each string, how many crossings, then

    - for each crossing, the cross name, then `1` if over, `-1` if under

- for each crossing, the name of the crossing and `1` if right, `-1` if left.

If there are `n` crossings, they must be named `0, 1, ..., n-1`.

AUTHORS:

- Miguel Marco (2015-03-24): initial version.
"""

# ****************************************************************************
#       Copyright (C) 2015 Miguel Marco  <mmarco@unizar.es>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at youroption) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from cysignals.signals cimport sig_on, sig_off

from sage.cpython.string cimport str_to_bytes, char_to_str

cdef extern from "homfly.h":
    ctypedef int  word
    ctypedef signed long int sb4
    ctypedef unsigned short int ub2
    ctypedef signed short int sb2
    struct Term:
        sb4 coef
        sb2 m
        sb2 l
    struct Poly:
        Term* term
        sb4 len
    Poly* homfly(char *argv)
    char* homfly_str(char *argv)


def homfly_polynomial_string(link):
    r"""
    Return the HOMFLY polynomial of a link.

    INPUT:

    - ``link`` -- string of space-separated integers representing the link

    OUTPUT: string with the HOMFLY polynomial in the variables `M` and `L`

    EXAMPLES::

        sage: from sage.libs.homfly import homfly_polynomial_string
        sage: trefoil = '1 6 0 1  1 -1  2 1  0 -1  1 1  2 -1 0 1 1 1 2 1'
        sage: homfly_polynomial_string(trefoil)
        ' - L^-4 - 2L^-2 + M^2L^-2'
    """
    link = str_to_bytes(link)
    sig_on()
    cdef char* c_output = homfly_str(link)
    sig_off()
    return char_to_str(c_output)


def homfly_polynomial_dict(link):
    """
    Return a dictionary representing the HOMFLY polynomial of a link.

    INPUT:

    - ``link`` -- string of space-separated integers representing the link

    EXAMPLES::

        sage: from sage.libs.homfly import homfly_polynomial_dict
        sage: trefoil = '1 6 0 1  1 -1  2 1  0 -1  1 1  2 -1 0 1 1 1 2 1'
        sage: homfly_polynomial_dict(trefoil)
        {(-4, 0): -1, (-2, 0): -2, (-2, 2): 1}
    """
    link = str_to_bytes(link)
    cdef Term ter
    sig_on()
    cdef Poly* c_output = homfly(link)
    sig_off()
    cdef int l = c_output.len
    d = {}
    for i in range(l):
        ter = c_output.term[i]
        d[(int(ter.l), int(ter.m))] = int(ter.coef)
    return d



# sage.doctest: needs sage.libs.pari
"""
Modular forms over a non-minimal base ring
"""

#########################################################################
#       Copyright (C) 2006 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#                  http://www.gnu.org/licenses/
#########################################################################

from . import ambient
from .cuspidal_submodule import CuspidalSubmodule_R
from sage.rings.integer_ring import ZZ
from sage.misc.cachefunc import cached_method


class ModularFormsAmbient_R(ambient.ModularFormsAmbient):
    def __init__(self, M, base_ring):
        """
        Ambient space of modular forms over a ring other than QQ.

        EXAMPLES::

            sage: M = ModularForms(23, 2, base_ring=GF(7)) # indirect doctest
            sage: M
            Modular Forms space of dimension 3 for Congruence Subgroup Gamma0(23)
             of weight 2 over Finite Field of size 7
            sage: M == loads(dumps(M))
            True
        """
        self.__M = M
        if M.character() is not None:
            self.__R_character = M.character().change_ring(base_ring)
        else:
            self.__R_character = None
        ambient.ModularFormsAmbient.__init__(self, M.group(), M.weight(), base_ring, M.character(), M._eis_only)

    @cached_method(key=lambda self, sign: ZZ(sign))  # convert sign to an Integer before looking this up in the cache
    def modular_symbols(self, sign=0):
        r"""
        Return the space of modular symbols attached to this space, with the given sign (default 0).

        TESTS::

            sage: # needs sage.rings.number_field
            sage: K.<i> = QuadraticField(-1)
            sage: chi = DirichletGroup(5, base_ring=K).0
            sage: x = polygen(ZZ, 'x')
            sage: L.<c> = K.extension(x^2 - 402*i)
            sage: M = ModularForms(chi, 7, base_ring=L)
            sage: symbs = M.modular_symbols()
            sage: symbs.character() == chi
            True
            sage: symbs.base_ring() == L
            True
        """
        sign = ZZ(sign)
        return self.__M.modular_symbols(sign).change_ring(self.base_ring())

    def _repr_(self):
        """
        String representation for ``self``.

        EXAMPLES::

            sage: M = ModularForms(23,2,base_ring=GF(7)) # indirect doctest
            sage: M._repr_()
            'Modular Forms space of dimension 3 for Congruence Subgroup Gamma0(23) of weight 2 over Finite Field of size 7'

            sage: # needs sage.rings.number_field
            sage: chi = DirichletGroup(109).0 ** 36
            sage: ModularForms(chi, 2, base_ring = chi.base_ring())
            Modular Forms space of dimension 9, character [zeta3] and weight 2 over Cyclotomic Field of order 108 and degree 36
        """
        s = str(self.__M)
        i = s.find('over')
        if i != -1:
            s = s[:i]
        return s + 'over %s' % self.base_ring()

    def _compute_q_expansion_basis(self, prec=None):
        """
        Compute `q`-expansions for a basis of ``self`` to precision ``prec``.

        EXAMPLES::

            sage: M = ModularForms(23,2,base_ring=GF(7))
            sage: M._compute_q_expansion_basis(10)
            [q + 6*q^3 + 6*q^4 + 5*q^6 + 2*q^7 + 6*q^8 + 2*q^9 + O(q^10),
             q^2 + 5*q^3 + 6*q^4 + 2*q^5 + q^6 + 2*q^7 + 5*q^8 + O(q^10),
             1 + 5*q^3 + 5*q^4 + 5*q^6 + 3*q^8 + 5*q^9 + O(q^10)]

        TESTS:

        This checks that :issue:`13445` is fixed::

            sage: M = ModularForms(Gamma0(11), base_ring=GF(11))
            sage: S = M.cuspidal_subspace()
            sage: 0 in [f.valuation() for f in S.basis()]
            False
            sage: from sage.modular.dims import dimension_cusp_forms
            sage: len(S.basis()) == dimension_cusp_forms(Gamma0(11), 2)
            True
        """
        if prec is None:
            prec = self.prec()
        R = self._q_expansion_ring()
        c = self.base_ring().characteristic()
        if c == 0:
            B = self.__M.q_expansion_basis(prec)
            return [R(f) for f in B]
        if c.is_prime_power():
            K = self.base_ring()
            p = K.characteristic().prime_factors()[0]
            from sage.rings.finite_rings.finite_field_constructor import GF
            Kp = GF(p)
            newB = [f.change_ring(K) for f in list(self.__M.cuspidal_subspace().q_integral_basis(prec))]
            A = Kp**prec
            gens = [f.padded_list(prec) for f in newB]
            V = A.span(gens)
            B = [f.change_ring(K) for f in self.__M.q_integral_basis(prec)]
            for f in B:
                fc = f.padded_list(prec)
                gens.append(fc)
                if not A.span(gens) == V:
                    newB.append(f)
                    V = A.span(gens)
            if len(newB) != self.dimension():
                raise RuntimeError("The dimension of the space is %s but the basis we computed has %s elements" % (self.dimension(), len(newB)))
            lst = [R(f) for f in newB]
            return [f/f[f.valuation()] for f in lst]
        # this returns a basis of q-expansions, without guaranteeing that
        # the first vectors form a basis of the cuspidal subspace
        # TODO: bring this in line with the other cases
        # simply using the above code fails because free modules over
        # general rings do not have a .span() method
        B = self.__M.q_integral_basis(prec)
        return [R(f) for f in B]

    def cuspidal_submodule(self):
        r"""
        Return the cuspidal subspace of this space.

        EXAMPLES::

            sage: # needs sage.rings.number_field
            sage: C = CuspForms(7, 4, base_ring=CyclotomicField(5))  # indirect doctest
            sage: type(C)
            <class 'sage.modular.modform.cuspidal_submodule.CuspidalSubmodule_R_with_category'>
        """
        return CuspidalSubmodule_R(self)

    def change_ring(self, R):
        r"""
        Return this modular forms space with the base ring changed to the ring R.

        EXAMPLES::

            sage: # needs sage.rings.number_field
            sage: chi = DirichletGroup(109, CyclotomicField(3)).0
            sage: M9 = ModularForms(chi, 2, base_ring = CyclotomicField(9))
            sage: M9.change_ring(CyclotomicField(15))
            Modular Forms space of dimension 10, character [zeta3 + 1] and weight 2
             over Cyclotomic Field of order 15 and degree 8
            sage: M9.change_ring(QQ)
            Traceback (most recent call last):
            ...
            ValueError: Space cannot be defined over Rational Field
        """
        if not R.has_coerce_map_from(self.__M.base_ring()):
            raise ValueError("Space cannot be defined over %s" % R)
        return ModularFormsAmbient_R(self.__M, R)



# distutils: libraries = flint
# distutils: depends = flint/aprcl.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/tools/flint-autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    bint aprcl_is_prime(const fmpz_t n) noexcept
    bint aprcl_is_prime_jacobi(const fmpz_t n) noexcept
    bint aprcl_is_prime_gauss(const fmpz_t n) noexcept
    primality_test_status _aprcl_is_prime_jacobi(const fmpz_t n, const aprcl_config config) noexcept
    primality_test_status _aprcl_is_prime_gauss(const fmpz_t n, const aprcl_config config) noexcept
    bint aprcl_is_prime_gauss_min_R(const fmpz_t n, ulong R) noexcept
    bint aprcl_is_prime_final_division(const fmpz_t n, const fmpz_t s, ulong r) noexcept
    void aprcl_config_gauss_init(aprcl_config conf, const fmpz_t n) noexcept
    void aprcl_config_gauss_init_min_R(aprcl_config conf, const fmpz_t n, ulong R) noexcept
    void aprcl_config_gauss_clear(aprcl_config conf) noexcept
    ulong aprcl_R_value(const fmpz_t n) noexcept
    void aprcl_config_jacobi_init(aprcl_config conf, const fmpz_t n) noexcept
    void aprcl_config_jacobi_clear(aprcl_config conf) noexcept
    void unity_zp_init(unity_zp f, ulong p, ulong exp, const fmpz_t n) noexcept
    void unity_zp_clear(unity_zp f) noexcept
    void unity_zp_copy(unity_zp f, const unity_zp g) noexcept
    void unity_zp_swap(unity_zp f, unity_zp q) noexcept
    void unity_zp_set_zero(unity_zp f) noexcept
    slong unity_zp_is_unity(unity_zp f) noexcept
    bint unity_zp_equal(unity_zp f, unity_zp g) noexcept
    void unity_zp_print(const unity_zp f) noexcept
    void unity_zp_coeff_set_fmpz(unity_zp f, ulong ind, const fmpz_t x) noexcept
    void unity_zp_coeff_set_ui(unity_zp f, ulong ind, ulong x) noexcept
    void unity_zp_coeff_add_fmpz(unity_zp f, ulong ind, const fmpz_t x) noexcept
    void unity_zp_coeff_add_ui(unity_zp f, ulong ind, ulong x) noexcept
    void unity_zp_coeff_inc(unity_zp f, ulong ind) noexcept
    void unity_zp_coeff_dec(unity_zp f, ulong ind) noexcept
    void unity_zp_mul_scalar_fmpz(unity_zp f, const unity_zp g, const fmpz_t s) noexcept
    void unity_zp_mul_scalar_ui(unity_zp f, const unity_zp g, ulong s) noexcept
    void unity_zp_add(unity_zp f, const unity_zp g, const unity_zp h) noexcept
    void unity_zp_mul(unity_zp f, const unity_zp g, const unity_zp h) noexcept
    void unity_zp_sqr(unity_zp f, const unity_zp g) noexcept
    void unity_zp_mul_inplace(unity_zp f, const unity_zp g, const unity_zp h, fmpz_t * t) noexcept
    void unity_zp_sqr_inplace(unity_zp f, const unity_zp g, fmpz_t * t) noexcept
    void unity_zp_pow_fmpz(unity_zp f, const unity_zp g, const fmpz_t pow) noexcept
    void unity_zp_pow_ui(unity_zp f, const unity_zp g, ulong pow) noexcept
    ulong _unity_zp_pow_select_k(const fmpz_t n) noexcept
    void unity_zp_pow_2k_fmpz(unity_zp f, const unity_zp g, const fmpz_t pow) noexcept
    void unity_zp_pow_2k_ui(unity_zp f, const unity_zp g, ulong pow) noexcept
    void unity_zp_pow_sliding_fmpz(unity_zp f, unity_zp g, const fmpz_t pow) noexcept
    void _unity_zp_reduce_cyclotomic_divmod(unity_zp f) noexcept
    void _unity_zp_reduce_cyclotomic(unity_zp f) noexcept
    void unity_zp_reduce_cyclotomic(unity_zp f, const unity_zp g) noexcept
    void unity_zp_aut(unity_zp f, const unity_zp g, ulong x) noexcept
    void unity_zp_aut_inv(unity_zp f, const unity_zp g, ulong x) noexcept
    void unity_zp_jacobi_sum_pq(unity_zp f, ulong q, ulong p) noexcept
    void unity_zp_jacobi_sum_2q_one(unity_zp f, ulong q) noexcept
    void unity_zp_jacobi_sum_2q_two(unity_zp f, ulong q) noexcept
    void unity_zpq_init(unity_zpq f, ulong q, ulong p, const fmpz_t n) noexcept
    void unity_zpq_clear(unity_zpq f) noexcept
    void unity_zpq_copy(unity_zpq f, const unity_zpq g) noexcept
    void unity_zpq_swap(unity_zpq f, unity_zpq q) noexcept
    bint unity_zpq_equal(const unity_zpq f, const unity_zpq g) noexcept
    slong unity_zpq_p_unity(const unity_zpq f) noexcept
    bint unity_zpq_is_p_unity(const unity_zpq f) noexcept
    bint unity_zpq_is_p_unity_generator(const unity_zpq f) noexcept
    void unity_zpq_coeff_set_fmpz(unity_zpq f, slong i, slong j, const fmpz_t x) noexcept
    void unity_zpq_coeff_set_ui(unity_zpq f, slong i, slong j, ulong x) noexcept
    void unity_zpq_coeff_add(unity_zpq f, slong i, slong j, const fmpz_t x) noexcept
    void unity_zpq_add(unity_zpq f, const unity_zpq g, const unity_zpq h) noexcept
    void unity_zpq_mul(unity_zpq f, const unity_zpq g, const unity_zpq h) noexcept
    void _unity_zpq_mul_unity_p(unity_zpq f) noexcept
    void unity_zpq_mul_unity_p_pow(unity_zpq f, const unity_zpq g, slong k) noexcept
    void unity_zpq_pow(unity_zpq f, const unity_zpq g, const fmpz_t p) noexcept
    void unity_zpq_pow_ui(unity_zpq f, const unity_zpq g, ulong p) noexcept
    void unity_zpq_gauss_sum(unity_zpq f, ulong q, ulong p) noexcept
    void unity_zpq_gauss_sum_sigma_pow(unity_zpq f, ulong q, ulong p) noexcept



# sage.doctest: needs sage.graphs
"""
GLPK Backend for access to GLPK graph functions

AUTHORS:

- Christian Kuper (2012-11): Initial implementation

Methods index
-------------

**Graph creation and modification operations:**

.. csv-table::
    :class: contentstable
    :widths: 30, 70
    :delim: |

    :meth:`~GLPKGraphBackend.add_vertex`          | Add an isolated vertex to the graph.
    :meth:`~GLPKGraphBackend.add_vertices`        | Add vertices from an iterable container of vertices.
    :meth:`~GLPKGraphBackend.set_vertex_demand`   | Set the vertex parameters.
    :meth:`~GLPKGraphBackend.set_vertices_demand` | Set the parameters of selected vertices.
    :meth:`~GLPKGraphBackend.get_vertex`          | Return a specific vertex as a dictionary Object.
    :meth:`~GLPKGraphBackend.get_vertices`        | Return a dictionary of the dictionaries associated to each vertex.
    :meth:`~GLPKGraphBackend.vertices`            | Return a ``list`` of all vertices.
    :meth:`~GLPKGraphBackend.delete_vertex`       | Remove a vertex from the graph.
    :meth:`~GLPKGraphBackend.delete_vertices`     | Remove vertices from the graph.
    :meth:`~GLPKGraphBackend.add_edge`            | Add an edge between vertices ``u`` and ``v``.
    :meth:`~GLPKGraphBackend.add_edges`           | Add edges to the graph.
    :meth:`~GLPKGraphBackend.get_edge`            | Return an edge connecting two vertices.
    :meth:`~GLPKGraphBackend.edges`               | Return a ``list`` of all edges in the graph.
    :meth:`~GLPKGraphBackend.delete_edge`         | Delete an edge from the graph.
    :meth:`~GLPKGraphBackend.delete_edges`        | Delete edges from the graph.

**Graph writing operations:**

.. csv-table::
    :class: contentstable
    :widths: 30, 70
    :delim: |

    :meth:`~GLPKGraphBackend.write_graph`    | Write the graph to a plain text file.
    :meth:`~GLPKGraphBackend.write_ccdata`   | Write the graph to a text file in DIMACS format.
    :meth:`~GLPKGraphBackend.write_mincost`  | Write the mincost flow problem data to a text file in DIMACS format.
    :meth:`~GLPKGraphBackend.write_maxflow`  | Write the maximum flow problem data to a text file in DIMACS format.

**Network optimization operations:**

.. csv-table::
    :class: contentstable
    :widths: 30, 70
    :delim: |

    :meth:`~GLPKGraphBackend.mincost_okalg`  | Find solution to the mincost problem with the out-of-kilter algorithm.
    :meth:`~GLPKGraphBackend.maxflow_ffalg`  | Find solution to the maxflow problem with Ford-Fulkerson algorithm.
    :meth:`~GLPKGraphBackend.cpp`            | Solve the critical path problem of a project network.

Classes and methods
-------------------
"""

#*****************************************************************************
#       Copyright (C) 2012 Christian Kuper <christian.kuper@t-online.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from cysignals.memory cimport check_allocarray, sig_free

from sage.cpython.string cimport str_to_bytes, char_to_str
from sage.cpython.string import FS_ENCODING
from sage.libs.glpk.constants cimport *
from sage.libs.glpk.graph cimport *
from sage.numerical.mip import MIPSolverException

cdef class GLPKGraphBackend():
    """
    GLPK Backend for access to GLPK graph functions.

    The constructor can either be called without arguments (which results in an
    empty graph) or with arguments to read graph data from a file.

    INPUT:

    - ``data`` -- a filename or a :class:`Graph` object

    - ``format`` -- when ``data`` is a filename, specifies the format of the
      data read from a file.  The ``format`` parameter is a string and can take
      values as described in the table below.

    **Format parameters:**

    .. list-table::
     :widths: 10 70

     * - ``plain``

       - Read data from a plain text file containing the following information:

           | nv na
           | i[1] j[1]
           | i[2] j[2]
           | . . .
           | i[na] j[na]

         where:

         * nv is the number of vertices (nodes);

         * na is the number of arcs;

         * i[k], k = 1, . . . , na, is the index of tail vertex of arc k;

         * j[k], k = 1, . . . , na, is the index of head vertex of arc k.


     * - ``dimacs``

       - Read data from a plain ASCII text file in DIMACS format.
         A description of the DIMACS format can be found at
         http://dimacs.rutgers.edu/Challenges/.

     * - ``mincost``

       - Reads the mincost flow problem data from a text file in DIMACS format

     * - ``maxflow``

       - Reads the maximum flow problem data from a text file in DIMACS format

    .. NOTE::

        When ``data`` is a :class:`Graph`, the following restrictions are
        applied.

            * vertices -- the value of the demand of each vertex (see
              :meth:`set_vertex_demand`) is obtained from the numerical
              value associated with the key "rhs" if it is a dictionary.

            * edges -- The edge values used in the algorithms are read from the
              edges labels (and left undefined if the edge labels are equal to
              ``None``). To be defined, the labels must be dictionary objects with
              keys "low", "cap" and "cost". See :meth:`get_edge` for details.

    EXAMPLES:

    The following example creates an empty graph::

        sage: from sage.numerical.backends.glpk_graph_backend import GLPKGraphBackend
        sage: gbe = GLPKGraphBackend()

    The following example creates an empty graph, adds some data, saves the data
    to a file and loads it::

        sage: from sage.numerical.backends.glpk_graph_backend import GLPKGraphBackend
        sage: gbe = GLPKGraphBackend()
        sage: gbe.add_vertices([None, None])
        ['0', '1']
        sage: a = gbe.add_edge('0', '1')
        sage: import tempfile
        sage: with tempfile.NamedTemporaryFile() as f:
        ....:     _ = gbe.write_graph(f.name)
        ....:     gbe1 = GLPKGraphBackend(f.name, "plain")
        Writing graph to ...
        4 lines were written
        Reading graph from ...
        Graph has 2 vertices and 1 edge
        3 lines were read

    The following example imports a Sage ``Graph`` and then uses it to solve a
    maxflow problem::

        sage: from sage.numerical.backends.glpk_graph_backend import GLPKGraphBackend
        sage: g = graphs.PappusGraph()
        sage: for ed in g.edges(sort=False):
        ....:     g.set_edge_label(ed[0], ed[1], {"cap":1})
        sage: gbe = GLPKGraphBackend(g)
        sage: gbe.maxflow_ffalg('1', '2')
        3.0
    """

    def __cinit__(self, data=None, format="plain"):
        """
        Constructor.

        The constructor can either be called without arguments creating an empty
        graph or with arguments to read graph data from a file or a Sage
        :class:`Graph`. See documentation of :class:`GLPKGraphBackend` for
        details.

        EXAMPLES::

            sage: from sage.numerical.backends.glpk_graph_backend import GLPKGraphBackend
            sage: gbe = GLPKGraphBackend()
        """

        from sage.graphs.graph import Graph

        self.graph = <glp_graph*> glp_create_graph(sizeof(c_v_data),
                       sizeof(c_a_data))

        if self.graph is NULL:
            raise MemoryError("Error allocating memory.")

        self.s = 1
        self.t = 1

        if isinstance(data, str):
            fname = str_to_bytes(data, FS_ENCODING, 'surrogateescape')
            res = 0
            if format == "plain":
                res = glp_read_graph(self.graph, fname)
            elif format == "dimacs":
                res = glp_read_ccdata(self.graph, 0, fname)
            elif format == "mincost":
                res = glp_read_mincost(self.graph, 0, 0, sizeof(double),
                   sizeof(double) + sizeof(double), fname)
            elif format == "maxflow":
                res = glp_read_maxflow(self.graph, &self.s, &self.t,
                   sizeof(double), fname)
            if res != 0:
                raise IOError("Could not read graph from file %s" % (fname))

        elif isinstance(data, Graph):
            self.__add_vertices_sage(data)
            self.__add_edges_sage(data)
        else:
            ValueError("Input data is not supported")

    cpdef add_vertex(self, name=None):
        """
        Add an isolated vertex to the graph.

        If the vertex already exists, nothing is done.

        INPUT:

        - ``name`` -- string of max 255 chars length. If no name is
          specified, then the vertex will be represented by the string
          representation of the ID of the vertex or - if this already exists -
          a string representation of the least integer not already representing
          a vertex.

        OUTPUT:

        If no ``name`` is passed as an argument, the new vertex name is
        returned. ``None`` otherwise.

        EXAMPLES::

            sage: from sage.numerical.backends.glpk_graph_backend import GLPKGraphBackend
            sage: gbe = GLPKGraphBackend()
            sage: gbe.add_vertex()
            '0'
            sage: gbe.add_vertex("2")
            sage: gbe.add_vertex()
            '1'
        """
        cdef int n
        cdef vn_t = 0

        if name is not None and self._find_vertex(name) >= 0:
            return None

        cdef int vn = glp_add_vertices(self.graph, 1)

        if name is not None:
            glp_set_vertex_name(self.graph, vn, str_to_bytes(name))
            return None

        else:
            s = str(vn - 1)
            n = self._find_vertex(s)

            # This is costly, but hopefully will not happen often.
            while n >= 0:
                vn_t += 1
                s = str(vn_t - 1)
                n = self._find_vertex(s)

            glp_set_vertex_name(self.graph, vn, str_to_bytes(s))
            return s

    cpdef __add_vertices_sage(self, g):
        """
        Add vertices to the GLPK Graph.

        This function is only used when importing a
        :class:`~sage.graphs.generic_graph.GenericGraph` object.

        EXAMPLES::

            sage: from sage.numerical.backends.glpk_graph_backend import GLPKGraphBackend
            sage: g = graphs.PappusGraph()
            sage: for ed in g.edges(sort=False):
            ....:     g.set_edge_label(ed[0], ed[1], {"cap":1})
            sage: gbe = GLPKGraphBackend(g)
            sage: gbe.maxflow_ffalg('1', '2')
            3.0
        """
        cdef int n
        cdef int i
        cdef double rhs
        cdef glp_vertex* vert

        verts = g.vertices(sort=True)
        n = len(verts)
        if n < 1:
            raise ValueError("Graph must contain vertices")

        glp_add_vertices(self.graph, n)

        for i in range(n):
            vert = self.graph.v[i+1]
            s = str(verts[i])
            glp_set_vertex_name(self.graph, i + 1, str_to_bytes(s))

            if g.get_vertex(verts[i]) is not None:
                try:
                    (<c_v_data *>vert.data).rhs = g.get_vertex(verts[i])["rhs"]
                except AttributeError:
                    pass

        glp_create_v_index(self.graph)

    cpdef list add_vertices(self, vertices):
        """
        Add vertices from an iterable container of vertices.

        Vertices that already exist in the graph will not be added again.

        INPUT:

        - ``vertices`` -- iterator of vertex labels (string); a label can be
          ``None``

        OUTPUT:

        Generated names of new vertices if there is at least one ``None`` value
        present in ``vertices``. ``None`` otherwise.

        EXAMPLES::

            sage: from sage.numerical.backends.glpk_graph_backend import GLPKGraphBackend
            sage: gbe = GLPKGraphBackend()
            sage: vertices = [None for i in range(3)]
            sage: gbe.add_vertices(vertices)
            ['0', '1', '2']
            sage: gbe.add_vertices(['A', 'B', None])
            ['5']
            sage: gbe.add_vertices(['A', 'B', 'C'])
            sage: gbe.vertices()
            ['0', '1', '2', 'A', 'B', '5', 'C']

        TESTS::

            sage: from sage.numerical.backends.glpk_graph_backend import GLPKGraphBackend
            sage: gbe = GLPKGraphBackend()
            sage: gbe.add_vertices([None, None, None, '1'])
            ['0', '2', '3']
        """

        # We do not want to have [None,None,None,1] as input as a vertex named
        # "1" would be created twice (a first time when adding a 'None' vertex,
        # and a second time when reading the last item of the list).
        nonecount = 0
        for v in vertices:
            if v is None:
                nonecount += 1
            else:
                self.add_vertex(v)

        if nonecount:
            return [self.add_vertex() for i in range(nonecount)]
        else:
            return None

    cpdef set_vertex_demand(self, vertex, demand):
        """
        Set the demand of the vertex in a mincost flow algorithm.

        INPUT:

        - ``vertex`` -- name of the vertex

        - ``demand`` -- the numerical value representing demand of the vertex in
          a mincost flow algorithm (it could be for instance `-1` to represent a
          sink, or `1` to represent a source and `0` for a neutral vertex). This
          can either be an ``int`` or ``float`` value.

        EXAMPLES::

            sage: from sage.numerical.backends.glpk_graph_backend import GLPKGraphBackend
            sage: gbe = GLPKGraphBackend()
            sage: vertices = [None for i in range(3)]
            sage: gbe.add_vertices(vertices)
            ['0', '1', '2']
            sage: gbe.set_vertex_demand('0', 2)
            sage: gbe.get_vertex('0')['rhs']
            2.0
            sage: gbe.set_vertex_demand('3', 2)
            Traceback (most recent call last):
            ...
            KeyError: 'Vertex 3 does not exist.'
        """
        cdef int n = self._find_vertex(vertex)

        if n < 0:
            raise KeyError("Vertex " + vertex + " does not exist.")

        cdef glp_vertex* vert = self.graph.v[n+1]
        cdef double val = demand
        (<c_v_data *>vert.data).rhs = val

    cpdef set_vertices_demand(self, list pairs):
        """
        Set the parameters of selected vertices.

        INPUT:

        - ``pairs`` -- list of pairs ``(vertex, demand)`` associating a demand
          to each vertex. For more information, see the documentation of
          :meth:`set_vertex_demand`.

        EXAMPLES::

            sage: from sage.numerical.backends.glpk_graph_backend import GLPKGraphBackend
            sage: gbe = GLPKGraphBackend()
            sage: vertices = [None for i in range(3)]
            sage: gbe.add_vertices(vertices)
            ['0', '1', '2']
            sage: gbe.set_vertices_demand([('0', 2), ('1', 3), ('3', 4)])
            sage: sorted(gbe.get_vertex('1').items())
            [('cut', 0), ('es', 0.0), ('ls', 0.0), ('pi', 0.0), ('rhs', 3.0)]
        """

        for v, param in pairs:
            try:
                self.set_vertex_demand(v, param)
            except KeyError:
                pass

    cpdef dict get_vertex(self, vertex):
        """
        Return a specific vertex as a dictionary Object.

        INPUT:

        - ``vertex`` -- the vertex label as string

        OUTPUT:

        The vertex as a dictionary object or ``None`` if the vertex does not
        exist. The dictionary contains the values used or created by the different
        algorithms. The values associated with the keys following keys contain:

            * "rhs"  -- The supply / demand value the vertex (mincost alg)
            * "pi"   -- The node potential (mincost alg)
            * "cut"  -- The cut flag of the vertex (maxflow alg)
            * "es"   -- The earliest start of task (cpp alg)
            * "ls"   -- The latest start of task (cpp alg)

        EXAMPLES::

            sage: from sage.numerical.backends.glpk_graph_backend import GLPKGraphBackend
            sage: gbe = GLPKGraphBackend()
            sage: verts = ["A", "B", "C", "D"]
            sage: gbe.add_vertices(verts)
            sage: sorted(gbe.get_vertex("A").items())
            [('cut', 0), ('es', 0.0), ('ls', 0.0), ('pi', 0.0), ('rhs', 0.0)]
            sage: gbe.get_vertex("F") is None
            True
        """

        cdef int i = self._find_vertex(vertex)
        if i < 0:
            return None

        cdef glp_vertex* vert = self.graph.v[i+1]
        cdef c_v_data * vdata = <c_v_data *> vert.data

        return {
            "rhs": vdata.rhs,
            "pi": vdata.pi,
            "cut": vdata.cut,
            "es": vdata.es,
            "ls": vdata.ls
            }

    cpdef dict get_vertices(self, verts):
        """
        Return a dictionary of the dictionaries associated to each vertex.

        INPUT:

        - ``verts`` -- iterable container of vertices

        OUTPUT:

        A list of pairs ``(vertex, properties)`` where ``properties`` is a
        dictionary containing the numerical values associated with a vertex. For
        more information, see the documentation of
        :meth:`GLPKGraphBackend.get_vertex`.

        EXAMPLES::

            sage: from sage.numerical.backends.glpk_graph_backend import GLPKGraphBackend
            sage: gbe = GLPKGraphBackend()
            sage: verts = ['A', 'B']
            sage: gbe.add_vertices(verts)
            sage: sorted(gbe.get_vertices(verts)['B'].items())
            [('cut', 0), ('es', 0.0), ('ls', 0.0), ('pi', 0.0), ('rhs', 0.0)]
            sage: gbe.get_vertices(["C", "D"])
            {}
        """
        vl = [(v, self.get_vertex(v)) for v in verts]
        return dict([(v, p) for v, p in vl if p is not None])

    cpdef list vertices(self):
        """
        Return the list of all vertices.

        .. NOTE::

           Changing elements of the ``list`` will not change anything in the
           the graph.

        .. NOTE::

           If a vertex in the graph does not have a name / label it will appear
           as ``None`` in the resulting ``list``.

        EXAMPLES::

            sage: from sage.numerical.backends.glpk_graph_backend import GLPKGraphBackend
            sage: gbe = GLPKGraphBackend()
            sage: verts = ["A", "B", "C"]
            sage: gbe.add_vertices(verts)
            sage: a = gbe.vertices(); a
            ['A', 'B', 'C']
            sage: a.pop(0)
            'A'
            sage: gbe.vertices()
            ['A', 'B', 'C']
        """

        return [char_to_str(self.graph.v[i+1].name)
                if self.graph.v[i+1].name is not NULL else None
                for i in range(self.graph.nv)]

    cpdef add_edge(self, u, v, dict params=None):
        """
        Add an edge between vertices ``u`` and ``v``.

        Allows adding an edge and optionally providing parameters used by the
        algorithms. If a vertex does not exist it is created.

        INPUT:

        - ``u`` -- the name (as string) of the tail vertex

        - ``v`` -- the name (as string) of the head vertex

        - ``params`` -- an optional dictionary containing the edge parameters used
          for the algorithms. The following keys are used:

            * ``low`` -- the minimum flow through the edge

            * ``cap`` -- the maximum capacity of the edge

            * ``cost`` -- the cost of transporting one unit through the edge

        EXAMPLES::

            sage: from sage.numerical.backends.glpk_graph_backend import GLPKGraphBackend
            sage: gbe = GLPKGraphBackend()
            sage: gbe.add_edge("A", "B", {"low":0.0, "cap":10.0, "cost":5})
            sage: gbe.vertices()
            ['A', 'B']
            sage: for ed in gbe.edges():
            ....:     print((ed[0], ed[1], ed[2]['cap'], ed[2]['cost'], ed[2]['low']))
            ('A', 'B', 10.0, 5.0, 0.0)
            sage: gbe.add_edge("B", "C", {"low":0.0, "cap":10.0, "cost":'5'})
            Traceback (most recent call last):
            ...
            TypeError: Invalid edge parameter.
        """
        cdef int i = self._find_vertex(u)
        cdef int j = self._find_vertex(v)

        if i < 0:
            self.add_vertex(u)
            i = self._find_vertex(u)

        if j < 0:
            self.add_vertex(v)
            j = self._find_vertex(v)

        cdef glp_arc *a

        a = glp_add_arc(self.graph, i+1, j+1)

        if params is not None:
            try:
                if "low" in params:
                    (<c_a_data *>a.data).low = params["low"]
                if "cap" in params:
                    (<c_a_data *>a.data).cap = params["cap"]
                if "cost" in params:
                    (<c_a_data *>a.data).cost = params["cost"]
            except TypeError:
                glp_del_arc(self.graph, a)
                raise TypeError("Invalid edge parameter.")

    cpdef list add_edges(self, edges):
        """
        Add edges to the graph.

        INPUT:

        - ``edges`` -- an iterable container of pairs of the form ``(u, v)``,
          where ``u`` is name (as string) of the tail vertex and ``v`` is the
          name (as string) of the head vertex or an iterable container of
          triples of the form ``(u, v, params)`` where params is a dictionary as
          described in ``add_edge``.

        EXAMPLES::

            sage: from sage.numerical.backends.glpk_graph_backend import GLPKGraphBackend
            sage: gbe = GLPKGraphBackend()
            sage: edges = [("A", "B", {"low":0.0, "cap":10.0, "cost":5})]
            sage: edges.append(("B", "C"))
            sage: gbe.add_edges(edges)
            sage: for ed in gbe.edges():
            ....:     print((ed[0], ed[1], ed[2]['cap'], ed[2]['cost'], ed[2]['low']))
            ('A', 'B', 10.0, 5.0, 0.0)
            ('B', 'C', 0.0, 0.0, 0.0)
            sage: edges = [("C", "D", {"low":0.0, "cap":10.0, "cost":5})]
            sage: edges.append(("C", "E", 5))
            sage: gbe.add_edges(edges)
            Traceback (most recent call last):
            ...
            TypeError: Argument 'params' has incorrect type ...
            sage: for ed in gbe.edges():
            ....:     print((ed[0], ed[1], ed[2]['cap'], ed[2]['cost'], ed[2]['low']))
            ('A', 'B', 10.0, 5.0, 0.0)
            ('B', 'C', 0.0, 0.0, 0.0)
            ('C', 'D', 10.0, 5.0, 0.0)
        """
        for ed in edges:
            self.add_edge(*ed)

    cpdef __add_edges_sage(self, g):
        """
        Add edges to the Graph.

        This function is only used when importing a ``GenericGraph``.

        EXAMPLES::

            sage: from sage.numerical.backends.glpk_graph_backend import GLPKGraphBackend
            sage: g = graphs.PappusGraph()
            sage: for ed in g.edges(sort=False):
            ....:     g.set_edge_label(ed[0], ed[1], {"cap":1})
            sage: gbe = GLPKGraphBackend(g)
            sage: gbe.maxflow_ffalg('1', '2')
            3.0
        """
        cdef glp_arc* a
        cdef int u
        cdef int v
        cdef double cost
        cdef double cap
        cdef double low
        cdef int isdirected = g.is_directed()

        for eu, ev, label in g.edges(sort=False):
            u_name = str(eu)
            v_name = str(ev)
            u = glp_find_vertex(self.graph, str_to_bytes(u_name))
            v = glp_find_vertex(self.graph, str_to_bytes(v_name))
            if u < 1 or v < 1:
                raise IndexError(u_name + " or " + v_name + " not found")

            a = glp_add_arc(self.graph, u, v)

            if isinstance(label, dict):
                if "cost" in label:
                    cost = label["cost"]
                    (<c_a_data *>a.data).cost = cost
                if "cap" in label:
                    cap = label["cap"]
                    (<c_a_data *>a.data).cap = cap
                if "low" in label:
                    low = label["low"]
                    (<c_a_data *>a.data).low = low

            if not isdirected:
                a = glp_add_arc(self.graph, v, u)
                if isinstance(label, dict):
                    if "cost" in label:
                        (<c_a_data *>a.data).cost = cost
                    if "cap" in label:
                        (<c_a_data *>a.data).cap = cap
                    if "low" in label:
                        (<c_a_data *>a.data).low = low

    cpdef tuple get_edge(self, u, v):
        """
        Return an edge connecting two vertices.

        .. NOTE::

           If multiple edges connect the two vertices only the first edge
           found is returned.

        INPUT:

        - ``u`` -- name (as string) of the tail vertex
        - ``v`` -- name (as string) of the head vertex

        OUTPUT:

        A ``triple`` describing if edge was found or ``None`` if not. The third
        value of the triple is a dictionary containing the following edge
        parameters:

            * ``low`` -- the minimum flow through the edge
            * ``cap`` -- the maximum capacity of the edge
            * ``cost`` -- the cost of transporting one unit through the edge
            * ``x`` -- the actual flow through the edge after solving

        EXAMPLES::

            sage: from sage.numerical.backends.glpk_graph_backend import GLPKGraphBackend
            sage: gbe = GLPKGraphBackend()
            sage: edges = [("A", "B"), ("A", "C"), ("B", "C")]
            sage: gbe.add_edges(edges)
            sage: ed = gbe.get_edge("A", "B")
            sage: ed[0], ed[1], ed[2]['x']
            ('A', 'B', 0.0)
            sage: gbe.get_edge("A", "F") is None
            True
        """
        cdef int i = self._find_vertex(u)
        cdef int j = self._find_vertex(v)

        if i < 0 or j < 0:
            return None

        cdef glp_vertex* vert_u = self.graph.v[i+1]
        cdef glp_vertex* vert_v = self.graph.v[j+1]
        cdef glp_arc* a = vert_u.out
        while a is not NULL:
            if a.head == vert_v:
                return (u, v, {"low":(<c_a_data *>a.data).low,
                               "cap":(<c_a_data *>a.data).cap,
                               "cost":(<c_a_data *>a.data).cost,
                               "x":(<c_a_data *>a.data).x})
            a = a.t_next

        return None

    cpdef list edges(self):
        """
        Return a ``list`` of all edges in the graph.

        OUTPUT: a ``list`` of ``triples`` representing the edges of the graph

        EXAMPLES::

            sage: from sage.numerical.backends.glpk_graph_backend import GLPKGraphBackend
            sage: gbe = GLPKGraphBackend()
            sage: edges = [("A", "B", {"low":0.0, "cap":10.0, "cost":5})]
            sage: edges.append(("B", "C"))
            sage: gbe.add_edges(edges)
            sage: for ed in gbe.edges():
            ....:     print((ed[0], ed[1], ed[2]['cost']))
            ('A', 'B', 5.0)
            ('B', 'C', 0.0)
        """

        cdef int i = 1
        cdef glp_vertex* vert_u
        cdef glp_vertex* vert_v
        cdef glp_arc* a
        edge_list = []

        while i <= self.graph.nv:
            vert_u = self.graph.v[i]
            a = vert_u.out
            while a is not NULL:
                vert_v = a.head
                if vert_u.name is NULL:
                    u_name = None
                else:
                    u_name = char_to_str(vert_u.name)
                if vert_v.name is NULL:
                    v_name = None
                else:
                    v_name = char_to_str(vert_v.name)
                edge_list.append((u_name, v_name,
                              {"low":(<c_a_data *>a.data).low,
                               "cap":(<c_a_data *>a.data).cap,
                               "cost":(<c_a_data *>a.data).cost,
                               "x":(<c_a_data *>a.data).x}))
                a = a.t_next
            i += 1
        return edge_list

    cpdef delete_vertex(self, vert):
        r"""
        Removes a vertex from the graph.

        Trying to delete a non existing vertex will raise an exception.

        INPUT:

        - ``vert`` -- the name (as string) of the vertex to delete

        EXAMPLES::

            sage: from sage.numerical.backends.glpk_graph_backend import GLPKGraphBackend
            sage: gbe = GLPKGraphBackend()
            sage: verts = ["A", "D"]
            sage: gbe.add_vertices(verts)
            sage: gbe.delete_vertex("A")
            sage: gbe.vertices()
            ['D']
            sage: gbe.delete_vertex("A")
            Traceback (most recent call last):
            ...
            RuntimeError: Vertex A does not exist.
        """

        cdef int i = self._find_vertex(vert)

        if i < 0:
            raise RuntimeError("Vertex %s does not exist." % vert)

        cdef int num[2]
        num[1] = i + 1
        cdef int ndel = 1

        glp_del_vertices(self.graph, ndel, num)

    cpdef delete_vertices(self, list verts):
        r"""
        Removes vertices from the graph.

        Trying to delete a non existing vertex will raise an exception.

        INPUT:

        - ``verts`` -- iterable container containing names (as string) of the
          vertices to delete

        EXAMPLES::

            sage: from sage.numerical.backends.glpk_graph_backend import GLPKGraphBackend
            sage: gbe = GLPKGraphBackend()
            sage: verts = ["A", "B", "C", "D"]
            sage: gbe.add_vertices(verts)
            sage: v_d = ["A", "B"]
            sage: gbe.delete_vertices(v_d)
            sage: gbe.vertices()
            ['C', 'D']
            sage: gbe.delete_vertices(["C", "A"])
            Traceback (most recent call last):
            ...
            RuntimeError: Vertex A does not exist.
            sage: gbe.vertices()
            ['C', 'D']
        """

        verts_val = [self._find_vertex(v) for v in verts]
        if -1 in verts_val:
            i = verts_val.index(-1)
            raise RuntimeError("Vertex %s does not exist." % verts[i])

        cdef int * num = <int *>check_allocarray(len(verts_val) + 1, sizeof(int))
        cdef int ndel = len(verts_val)

        for i,(v) in enumerate(verts_val):
            num[i+1] = v+1

        glp_del_vertices(self.graph, ndel, num)

        sig_free(num)

    cpdef delete_edge(self, u, v, dict params=None):
        """
        Delete an edge from the graph.

        If an edge does not exist it is ignored.

        INPUT:

        - ``u`` -- the name (as string) of the tail vertex of the edge
        - ``v`` -- the name (as string) of the tail vertex of the edge
        - ``params`` -- an optional dictionary containing the edge
          parameters (see :meth:`add_edge`). If this parameter
          is not provided, all edges connecting ``u`` and ``v`` are deleted.
          Otherwise only edges with matching parameters are deleted.

        .. SEEALSO::

            :meth:`delete_edges`

        EXAMPLES::

            sage: from sage.numerical.backends.glpk_graph_backend import GLPKGraphBackend
            sage: gbe = GLPKGraphBackend()
            sage: edges = [("A", "B", {"low":0.0, "cap":10.0, "cost":5})]
            sage: edges.append(("A", "B", {"low":0.0, "cap":15.0, "cost":10}))
            sage: edges.append(("B", "C", {"low":0.0, "cap":20.0, "cost":1}))
            sage: edges.append(("B", "C", {"low":0.0, "cap":10.0, "cost":20}))
            sage: gbe.add_edges(edges)
            sage: gbe.delete_edge("A", "B")
            sage: gbe.delete_edge("B", "C", {"low":0.0, "cap":10.0, "cost":20})
            sage: gbe.edges()[0][0], gbe.edges()[0][1], gbe.edges()[0][2]['cost']
            ('B', 'C', 1.0)
        """

        cdef int i = self._find_vertex(u)
        cdef int j = self._find_vertex(v)
        if i < 0 or j < 0:
            return

        cdef glp_vertex* vert_u = self.graph.v[i+1]
        cdef glp_vertex* vert_v = self.graph.v[j+1]
        cdef glp_arc* a = vert_u.out
        cdef glp_arc* a2 = a

        cdef double low, cap, cost, x

        if params is not None:
            if "low" in params:
                low = params["low"]
            if "cap" in params:
                cap = params["cap"]
            if "cost" in params:
                cost = params["cost"]
            if "x" in params:
                x = params["x"]

        while a is not NULL:
            a2 = a.t_next
            if a.head == vert_v and params is None:
                glp_del_arc(self.graph, a)
            elif a.head == vert_v:
                del_it = True
                if "low" in params:
                    if (<c_a_data *>a.data).low != low:
                        del_it = False
                if "cap" in params:
                    if (<c_a_data *>a.data).cap != cap:
                        del_it = False
                if "cost" in params:
                    if (<c_a_data *>a.data).cost != cost:
                        del_it = False
                if "x" in params:
                    if (<c_a_data *>a.data).x != x:
                        del_it = False
                if del_it:
                    glp_del_arc(self.graph, a)

            a = a2

    def delete_edges(self, edges):
        """
        Delete edges from the graph.

        Non existing edges are ignored.

        INPUT:

        - ``edges`` -- an iterable container of edges

        .. SEEALSO::

            :meth:`delete_edge`

        EXAMPLES::

            sage: from sage.numerical.backends.glpk_graph_backend import GLPKGraphBackend
            sage: gbe = GLPKGraphBackend()
            sage: edges = [("A", "B", {"low":0.0, "cap":10.0, "cost":5})]
            sage: edges.append(("A", "B", {"low":0.0, "cap":15.0, "cost":10}))
            sage: edges.append(("B", "C", {"low":0.0, "cap":20.0, "cost":1}))
            sage: edges.append(("B", "C", {"low":0.0, "cap":10.0, "cost":20}))
            sage: gbe.add_edges(edges)
            sage: gbe.delete_edges(edges[1:])
            sage: len(gbe.edges())
            1
            sage: gbe.edges()[0][0], gbe.edges()[0][1], gbe.edges()[0][2]['cap']
            ('A', 'B', 10.0)
        """

        for edge in edges:
            self.delete_edge(*edge)

    cpdef int _find_vertex(self, name) noexcept:
        """
        Return the index of a vertex specified by a name

        INPUT:

        - ``name`` -- name of the vertex

        OUTPUT: the index of the vertex or ``-1`` if the vertex is not found

        EXAMPLES::

            sage: from sage.numerical.backends.glpk_graph_backend import GLPKGraphBackend
            sage: gbe = GLPKGraphBackend()
            sage: verts = ["A", "B", "C", "D"]
            sage: gbe.add_vertices(verts)
            sage: gbe._find_vertex("A")
            0
            sage: gbe._find_vertex("F")
            -1
        """

        glp_create_v_index(self.graph)
        return glp_find_vertex(self.graph, str_to_bytes(name)) - 1

    cpdef int write_graph(self, fname) noexcept:
        r"""
        Writes the graph to a plain text file

        INPUT:

        - ``fname`` -- full name of the file

        OUTPUT: zero if the operations was successful otherwise nonzero

        EXAMPLES::

            sage: from sage.numerical.backends.glpk_graph_backend import GLPKGraphBackend
            sage: gbe = GLPKGraphBackend()
            sage: a = gbe.add_edge("0", "1")
            sage: import tempfile
            sage: with tempfile.NamedTemporaryFile() as f:
            ....:     gbe.write_graph(f.name)
            Writing graph to ...
            4 lines were written
            0
        """

        fname = str_to_bytes(fname, FS_ENCODING, 'surrogateescape')
        return glp_write_graph(self.graph, fname)

    cpdef int write_ccdata(self, fname) noexcept:
        r"""
        Writes the graph to a text file in DIMACS format.

        Writes the data to plain ASCII text file in DIMACS format.
        A description of the DIMACS format can be found at
        http://dimacs.rutgers.edu/Challenges/.

        INPUT:

        - ``fname`` -- full name of the file

        OUTPUT: zero if the operations was successful otherwise nonzero

        EXAMPLES::

            sage: from sage.numerical.backends.glpk_graph_backend import GLPKGraphBackend
            sage: gbe = GLPKGraphBackend()
            sage: a = gbe.add_edge("0", "1")
            sage: import tempfile
            sage: with tempfile.NamedTemporaryFile() as f:
            ....:     gbe.write_ccdata(f.name)
            Writing graph to ...
            6 lines were written
            0
        """

        fname = str_to_bytes(fname, FS_ENCODING, 'surrogateescape')
        return glp_write_ccdata(self.graph, 0, fname)

    cpdef int write_mincost(self, fname) noexcept:
        """
        Writes the mincost flow problem data to a text file in DIMACS format

        INPUT:

        - ``fname`` -- full name of file

        OUTPUT: zero if successful, otherwise nonzero

        EXAMPLES::

            sage: from sage.numerical.backends.glpk_graph_backend import GLPKGraphBackend
            sage: gbe = GLPKGraphBackend()
            sage: a = gbe.add_edge("0", "1")
            sage: import tempfile
            sage: with tempfile.NamedTemporaryFile() as f:
            ....:     gbe.write_mincost(f.name)
            Writing min-cost flow problem data to ...
            4 lines were written
            0
        """

        fname = str_to_bytes(fname, FS_ENCODING, 'surrogateescape')
        return glp_write_mincost(self.graph, 0, 0, sizeof(double),
                   sizeof(double) + sizeof(double), fname)

    cpdef double mincost_okalg(self) except -1:
        r"""
        Finds solution to the mincost problem with the out-of-kilter algorithm.

        The out-of-kilter algorithm requires all problem data to be integer
        valued.

        OUTPUT:

        The solution to the mincost problem, i.e. the total cost, if operation
        was successful.

        .. NOTE::

           This method raises ``MIPSolverException`` exceptions when
           the solution cannot be computed for any reason (none
           exists, or the LP solver was not able to find it, etc...)

        EXAMPLES::

            sage: from sage.numerical.backends.glpk_graph_backend import GLPKGraphBackend
            sage: gbe = GLPKGraphBackend()
            sage: vertices = (35, 50, 40, -45, -20, -30, -30)
            sage: vs = gbe.add_vertices([None for i in range(len(vertices))])
            sage: v_dict = {}
            sage: for i, v in enumerate(vs):
            ....:     v_dict[v] = vertices[i]
            sage: gbe.set_vertices_demand(list(v_dict.items()))
            sage: cost = ((8, 6, 10, 9), (9, 12, 13, 7), (14, 9, 16, 5))

            sage: for i in range(len(cost)):
            ....:     for j in range(len(cost[0])):
            ....:         gbe.add_edge(str(i), str(j + len(cost)), {"cost":cost[i][j], "cap":100})
            sage: gbe.mincost_okalg()
            1020.0
            sage: for ed in gbe.edges():
            ....:     print("{} -> {} {}".format(ed[0], ed[1], ed[2]["x"]))
            0 -> 6 0.0
            0 -> 5 25.0
            0 -> 4 10.0
            0 -> 3 0.0
            1 -> 6 0.0
            1 -> 5 5.0
            1 -> 4 0.0
            1 -> 3 45.0
            2 -> 6 30.0
            2 -> 5 0.0
            2 -> 4 10.0
            2 -> 3 0.0
        """
        cdef double graph_sol
        cdef int status = glp_mincost_okalg(self.graph, 0, 0, sizeof(double),
              2 * sizeof(double), &graph_sol,
              3 * sizeof(double), sizeof(double))
        if status == 0:
            pass
        elif status == GLP_ENOPFS:
            raise MIPSolverException("No (primal) feasible solution exists")
        elif status == GLP_EDATA:
            raise MIPSolverException("Unable to start search due to " +
                                     "problem data")
        elif status == GLP_ERANGE:
            raise MIPSolverException("The search was prematurely terminated " +
                                     "because of integer overflow")
        elif status == GLP_EFAIL:
            raise MIPSolverException("An error has been detected" +
                                     "in the program logic")

        return graph_sol

    cpdef int write_maxflow(self, fname) except -1:
        """
        Writes the maximum flow problem data to a text file in DIMACS format.

        INPUT:

        - ``fname`` -- full name of file

        OUTPUT:

        ``Zero`` if successful, otherwise ``nonzero``

        EXAMPLES::

            sage: from sage.numerical.backends.glpk_graph_backend import GLPKGraphBackend
            sage: gbe = GLPKGraphBackend()
            sage: import tempfile
            sage: with tempfile.NamedTemporaryFile() as f:
            ....:     gbe.write_maxflow(f.name)
            Traceback (most recent call last):
            ...
            OSError: Cannot write empty graph
            sage: gbe.add_vertices([None for i in range(2)])
            ['0', '1']
            sage: a = gbe.add_edge('0', '1')
            sage: gbe.maxflow_ffalg('0', '1')
            0.0
            sage: with tempfile.NamedTemporaryFile() as f:
            ....:     gbe.write_maxflow(f.name)
            Writing maximum flow problem data to ...
            6 lines were written
            0
        """

        if self.graph.nv <= 0:
            raise IOError("Cannot write empty graph")

        fname = str_to_bytes(fname, FS_ENCODING, 'surrogateescape')
        return glp_write_maxflow(self.graph, self.s+1, self.t+1,
                   sizeof(double), fname)

    cpdef double maxflow_ffalg(self, u=None, v=None) except -1:
        r"""
        Finds solution to the maxflow problem with Ford-Fulkerson algorithm.

        INPUT:

        - ``u`` -- name (as string) of the tail vertex; default is ``None``
        - ``v`` -- name (as string) of the head vertex; default is ``None``

        If ``u`` or ``v`` are ``None``, the currently stored values for the
        head or tail vertex are used. This behavior is useful when reading
        maxflow data from a file. When calling this function with values for
        ``u`` and ``v``, the head and tail vertex are stored for
        later use.

        OUTPUT: the solution to the maxflow problem, i.e. the maximum flow

        .. NOTE::

            * If the source or sink vertex does not exist, an :exc:`IndexError` is
              raised.

            * If the source and sink are identical, a :exc:`ValueError` is raised.

            * This method raises ``MIPSolverException`` exceptions when the
              solution cannot be computed for any reason (none exists, or the
              LP solver was not able to find it, etc...)

        EXAMPLES::

            sage: from sage.numerical.backends.glpk_graph_backend import GLPKGraphBackend
            sage: gbe = GLPKGraphBackend()
            sage: v = gbe.add_vertices([None for i in range(5)])
            sage: edges = ((0, 1, 2), (0, 2, 3), (1, 2, 3), (1, 3, 4),
            ....:         (3, 4, 1), (2, 4, 2))
            sage: for a in edges:
            ....:     edge = gbe.add_edge(str(a[0]), str(a[1]), {"cap":a[2]})
            sage: gbe.maxflow_ffalg('0', '4')
            3.0
            sage: gbe.maxflow_ffalg()
            3.0
            sage: gbe.maxflow_ffalg('0', '8')
            Traceback (most recent call last):
            ...
            IndexError: Source or sink vertex does not exist
        """
        cdef int s, t

        if u is not None and v is not None:
            s = self._find_vertex(u)
            t = self._find_vertex(v)
        else:
            s = self.s
            t = self.t

        if s < 0 or t < 0:
            raise IndexError("Source or sink vertex does not exist")
        if s == t:
            raise ValueError("Source and sink are identical")

        self.s = s
        self.t = t

        s += 1
        t += 1

        cdef double graph_sol
        cdef int status = glp_maxflow_ffalg(self.graph, s, t, sizeof(double),
                          &graph_sol, 3 * sizeof(double),
                          4 * sizeof(double))
        if status == 0:
            pass
        elif status == GLP_ENOPFS:
            raise MIPSolverException("No (primal) feasible solution exists")
        elif status == GLP_EDATA:
            raise MIPSolverException("Unable to start search due " +
                                     "to problem data")
        elif status == GLP_ERANGE:
            raise MIPSolverException("The search was prematurely terminated " +
                                     "because of integer overflow")
        elif status == GLP_EFAIL:
            raise MIPSolverException("An error has been detected in the " +
                                     "program logic")

        return graph_sol

    cpdef double cpp(self) noexcept:
        r"""
        Solve the critical path problem of a project network.

        OUTPUT: the length of the critical path of the network

        EXAMPLES::

            sage: from sage.numerical.backends.glpk_graph_backend import GLPKGraphBackend
            sage: gbe = GLPKGraphBackend()
            sage: gbe.add_vertices([None for i in range(3)])
            ['0', '1', '2']
            sage: gbe.set_vertex_demand('0', 3)
            sage: gbe.set_vertex_demand('1', 1)
            sage: gbe.set_vertex_demand('2', 4)
            sage: a = gbe.add_edge('0', '2')
            sage: a = gbe.add_edge('1', '2')
            sage: gbe.cpp()
            7.0
            sage: v = gbe.get_vertex('1')
            sage: 1, v["rhs"], v["es"], v["ls"] # abs tol 1e-6
            (1, 1.0, 0.0, 2.0)
        """

        return glp_cpp(self.graph, 0, 2 * sizeof(double),
                   3 * sizeof(double))

    def __dealloc__(self):
        """
        Destructor
        """
        if self.graph is not NULL:
            glp_delete_graph(self.graph)
        self.graph = NULL



r"""
Topological Closures of Manifold Subsets

:class:`ManifoldSubsetClosure` implements the topological closure
of a manifold subset in the topology of the manifold.
"""

# ****************************************************************************
#       Copyright (C) 2021 Matthias Koeppe <mkoeppe@math.ucdavis.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.manifolds.subset import ManifoldSubset


class ManifoldSubsetClosure(ManifoldSubset):
    r"""
    Topological closure of a manifold subset in the topology of the manifold.

    INPUT:

    - ``subset`` -- a :class:`~sage.manifolds.subset.ManifoldSubset`
    - ``name`` -- (default: computed from the name of the subset)
      string; name (symbol) given to the closure
    - ``latex_name`` -- string (default: ``None``); LaTeX symbol to
      denote the subset. If none is provided, it is set to ``name``.

    EXAMPLES::

        sage: M = Manifold(2, 'R^2', structure='topological')
        sage: c_cart.<x,y> = M.chart() # Cartesian coordinates on R^2
        sage: D = M.open_subset('D', coord_def={c_cart: x^2+y^2<1}); D
        Open subset D of the 2-dimensional topological manifold R^2
        sage: cl_D = D.closure()
        sage: cl_D
        Topological closure cl_D of the Open subset D of the 2-dimensional
         topological manifold R^2
        sage: latex(cl_D)
        \mathop{\mathrm{cl}}(D)
        sage: type(cl_D)
        <class 'sage.manifolds.subsets.closure.ManifoldSubsetClosure_with_category'>
        sage: cl_D.category()
        Category of subobjects of sets

    The closure of the subset `D` is a subset of every closed superset
    of `D`::

        sage: S = D.superset('S')
        sage: S.declare_closed()
        sage: cl_D.is_subset(S)
        True
    """

    def __init__(self, subset, name=None, latex_name=None):
        r"""
        Initialize a :class:`ManifoldSubsetClosure` instance.

        TESTS::

            sage: from sage.manifolds.subsets.closure import ManifoldSubsetClosure
            sage: M = Manifold(2, 'R^2', structure='topological')
            sage: c_cart.<x,y> = M.chart() # Cartesian coordinates on R^2
            sage: D = M.open_subset('D', coord_def={c_cart: x^2+y^2<1}); D
            Open subset D of the 2-dimensional topological manifold R^2
            sage: cl_D = D.closure(); cl_D  # indirect doctest
            Topological closure cl_D of the
             Open subset D of the 2-dimensional topological manifold R^2
            sage: also_cl_D = ManifoldSubsetClosure(D, name='also_cl_D'); also_cl_D
            Topological closure also_cl_D of the
             Open subset D of the 2-dimensional topological manifold R^2
            sage: cl_D is also_cl_D
            False
            sage: cl_D == also_cl_D
            False
        """
        self._subset = subset
        base_manifold = subset.manifold()
        if latex_name is None:
            if name is None:
                latex_name = r'\mathop{\mathrm{cl}}(' + subset._latex_name + ')'
            else:
                latex_name = name
        if name is None:
            name = 'cl_' + subset._name
        ManifoldSubset.__init__(self, base_manifold, name, latex_name=latex_name)
        self.declare_superset(subset)
        self.declare_subset(superset
                            for superset in subset.supersets()
                            if superset.is_closed())

    def _repr_(self):
        r"""
        String representation of the object.

        TESTS::

            sage: from sage.manifolds.subsets.closure import ManifoldSubsetClosure
            sage: M = Manifold(2, 'R^2', structure='topological')
            sage: D = M.open_subset('D')
            sage: cl_D = D.closure(); cl_D  # indirect doctest
            Topological closure cl_D of the
             Open subset D of the 2-dimensional topological manifold R^2
        """
        return "Topological closure {} of the {}".format(self._name, self._subset)

    def is_closed(self):
        r"""
        Return if ``self`` is a closed set.

        This implementation of the method always returns ``True``.

        EXAMPLES::

            sage: from sage.manifolds.subsets.closure import ManifoldSubsetClosure
            sage: M = Manifold(2, 'R^2', structure='topological')
            sage: c_cart.<x,y> = M.chart() # Cartesian coordinates on R^2
            sage: D = M.open_subset('D', coord_def={c_cart: x^2+y^2<1}); D
            Open subset D of the 2-dimensional topological manifold R^2
            sage: cl_D = D.closure(); cl_D  # indirect doctest
            Topological closure cl_D of the Open subset D of the 2-dimensional topological manifold R^2
            sage: cl_D.is_closed()
            True
        """
        return True



"""
`L`-functions from PARI

This is a wrapper around the general PARI `L`-functions functionality.

REFERENCES:

- https://pari.math.u-bordeaux.fr/dochtml/html-stable/_L_minusfunctions.html

AUTHORS:

- Frédéric Chapoton (2018) interface
"""
# ****************************************************************************
#       Copyright (C) 2018 Frédéric Chapoton <chapoton@unistra.fr>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from operator import index as PyNumber_Index

from cypari2.gen import Gen

from sage.libs.pari import pari
from sage.rings.integer_ring import ZZ
from sage.rings.real_mpfr import RealField
from sage.rings.complex_mpfr import ComplexField
from sage.rings.power_series_ring import PowerSeriesRing
from sage.structure.sage_object import SageObject


class lfun_generic:
    r"""
    Create a PARI `L`-function (:pari:`lfun` instance).

    The arguments are::

        lfun_generic(conductor, gammaV, weight, eps, poles, residues, init)

    where

    - ``conductor`` -- integer; the conductor

    - ``gammaV`` -- list of Gamma-factor parameters, e.g. [0] for
      Riemann zeta, [0,1] for elliptic curves, (see examples)

    - ``weight`` -- positive real number, usually an integer e.g. 1 for
      Riemann zeta, 2 for `H^1` of curves over `\QQ`

    - ``eps`` -- complex number; sign in functional equation

    - ``poles`` -- (default: ``[]``) list of points where
      `L^*(s)` has (simple) poles; only poles with
      `Re(s)>weight/2` should be included

    - ``residues`` -- vector of residues of `L^*(s)` in those poles or
      set ``residues='automatic'`` (default)

    - ``init`` -- list of coefficients (optional)

    RIEMANN ZETA FUNCTION:

    We compute with the Riemann Zeta function::

        sage: from sage.lfunctions.pari import lfun_generic, LFunction
        sage: lf = lfun_generic(conductor=1, gammaV=[0], weight=1,
        ....:                   eps=1, poles=[1], residues=[1])
        sage: lf.init_coeffs([1]*2000)

    Now we can wrap this PARI `L`-function into one Sage `L`-function::

        sage: L = LFunction(lf); L
        L-series of conductor 1 and weight 1
        sage: L(1)
        Traceback (most recent call last):
        ...
        ArithmeticError: pole here
        sage: L(2)
        1.64493406684823
        sage: L.derivative(2)
        -0.937548254315844
        sage: h = RR('0.0000000000001')
        sage: (zeta(2+h) - zeta(2.))/h
        -0.937028232783632
        sage: L.taylor_series(2, k=5)
        1.64493406684823 - 0.937548254315844*z + 0.994640117149451*z^2 - 1.00002430047384*z^3 + 1.00006193307...*z^4 + O(z^5)
    """
    def __init__(self, conductor, gammaV, weight, eps, poles=[],
                 residues='automatic', *args, **kwds) -> None:
        """
        Initialisation of a :pari:`lfun` from motivic data.

        This can happen either in one stage or in two stages: the coefficients
        can be given using the ``init`` keyword, or entered using
        the :meth:`init_coeffs`.

        EXAMPLES::

            sage: from sage.lfunctions.pari import lfun_generic, LFunction
            sage: lf = lfun_generic(conductor=1, gammaV=[0], weight=1,
            ....:                   eps=1, poles=[1], residues=[1])
            sage: lf
            generic Pari lfun structure of conductor 1 and weight 1
        """
        self.conductor = conductor
        self.gammaV = gammaV
        self.weight = weight
        self.eps = eps
        self.poles = poles
        self.residues = residues

        if (isinstance(self.poles, (list, tuple)) and
                isinstance(self.residues, (list, tuple))):
            if len(self.poles) != len(self.residues):
                raise ValueError("poles and residues do not match")

        if args or kwds:
            self.init_coeffs(*args, **kwds)
        else:
            # empty placeholder
            # useful to store parameters before computing the coefficients
            self.init_empty()

    def __repr__(self) -> str:
        """
        Return a string representation.

        EXAMPLES::

            sage: from sage.lfunctions.pari import lfun_generic, LFunction
            sage: lf = lfun_generic(conductor=1, gammaV=[0], weight=1,
            ....:                   eps=1, poles=[1], residues=[1])
            sage: lf
            generic Pari lfun structure of conductor 1 and weight 1
        """
        t = "generic Pari lfun structure "
        t += f"of conductor {self.conductor} and weight {self.weight}"
        return t

    def init_empty(self) -> None:
        """
        Create the Pari object with the Dokchitser parameters only.

        This is useful to ask Pari for the number of terms.

        EXAMPLES::

            sage: from sage.lfunctions.pari import lfun_generic, LFunction
            sage: lf = lfun_generic(conductor=1, gammaV=[0], weight=1,
            ....:                   eps=1, poles=[1], residues=[1])
            sage: lf.init_empty()
        """
        # empty placeholder
        # just storing the parameters, not the coefficients
        self._L = pari.lfuncreate([[], [], self.gammaV, self.weight,
                                   self.conductor, self.eps])

    def init_coeffs(self, v, w=1):
        """
        Set the coefficients `a_n` of the `L`-series.

        If `L(s)` is not equal to its dual, pass the coefficients of
        the dual as the second argument.

        INPUT:

        - ``v`` -- list of complex numbers or unary function

        - ``w`` -- list of complex numbers or unary function

        EXAMPLES::

            sage: from sage.lfunctions.pari import lfun_generic, LFunction
            sage: lf = lfun_generic(conductor=1, gammaV=[0, 1], weight=12, eps=1)
            sage: pari_coeffs = pari('k->vector(k,n,(5*sigma(n,3)+7*sigma(n,5))*n/12 - 35*sum(k=1,n-1,(6*k-4*(n-k))*sigma(k,3)*sigma(n-k,5)))')
            sage: lf.init_coeffs(pari_coeffs)

        Evaluate the resulting `L`-function at a point, and compare with
        the answer that one gets "by definition" (of `L`-function
        attached to a modular form)::

            sage: L = LFunction(lf)
            sage: L(14)
            0.998583063162746
            sage: a = delta_qexp(1000)
            sage: sum(a[n]/float(n)^14 for n in reversed(range(1,1000)))
            0.9985830631627461

        Illustrate that one can give a list of complex numbers for v
        (see :issue:`10937`)::

            sage: l2 = lfun_generic(conductor=1, gammaV=[0, 1], weight=12, eps=1)
            sage: l2.init_coeffs(list(delta_qexp(1000))[1:])
            sage: L2 = LFunction(l2)
            sage: L2(14)
            0.998583063162746

        TESTS:

        Verify that setting the `w` parameter does not raise an error
        (see :issue:`10937`)::

            sage: L2 = lfun_generic(conductor=1, gammaV=[0, 1], weight=12, eps=1)
            sage: L2.init_coeffs(list(delta_qexp(1000))[1:], w=[1..1000])
        """
        pv = pari(v)
        if pv.type() not in ('t_CLOSURE', 't_VEC'):
            raise TypeError("v (coefficients) must be a list or a function")

        # w = 0 means a*_n = a_n
        # w = 1 means a*_n = complex conjugate of a_n
        # otherwise w must be a list of coefficients
        pw = pari(w)
        if pw.type() not in ('t_INT', 't_CLOSURE', 't_VEC'):
            raise TypeError("w (dual coefficients) must be a list or a function or the special value 0 or 1")

        if isinstance(self.poles, (tuple, list)) and not self.poles:
            self._L = pari.lfuncreate([pv, pw, self.gammaV, self.weight,
                                       self.conductor, self.eps])
        elif self.poles == 0:
            # trying automatic pole reconstruction
            self._L = pari.lfuncreate([pv, pw, self.gammaV, self.weight,
                                       self.conductor, self.eps, 0])
        elif isinstance(self.residues, (list, tuple)):
            # pari expects pairs (pole, polar part as power series),
            # not residues
            # moreover, they must be the poles and polar parts of the
            # completed L-function (as t_Col)

            # it could also be those of the bare L-function (as t_Vec)
            # but we do not yet allow this
            x = pari("x")
            residues = (pari.Ser([r], "x", 1) / x for r in self.residues)
            poles = tuple(pari.Col([b, Pb])
                          for b, Pb in zip(self.poles, residues))
            self._L = pari.lfuncreate([pv, pw, self.gammaV, self.weight,
                                       self.conductor, self.eps,
                                       poles])
        else:
            # assuming a single pole, given as a complex scalar
            self._L = pari.lfuncreate([pv, pw, self.gammaV, self.weight,
                                       self.conductor, self.eps,
                                       self.poles[0]])

    def __pari__(self):
        """
        Return the PARI `L`-function object.

        EXAMPLES::

            sage: from sage.lfunctions.pari import lfun_generic, LFunction
            sage: lf = lfun_generic(conductor=1, gammaV=[0], weight=1, eps=1,
            ....:                   poles=[1], residues=[1])
            sage: lf.__pari__()
            [[Vecsmall([0]), []], [Vecsmall([0]), []], [0], 1, 1, 1]

            sage: lf.init_coeffs([1]*2000)
            sage: X = lf.__pari__()
            sage: X.type()
            't_VEC'
        """
        return self._L


def lfun_character(chi):
    """
    Create the `L`-function of a primitive Dirichlet character.

    If the given character is not primitive, it is replaced by its
    associated primitive character.

    OUTPUT: one :pari:`lfun` object

    EXAMPLES::

        sage: from sage.lfunctions.pari import lfun_character, LFunction
        sage: chi = DirichletGroup(6).gen().primitive_character()
        sage: L = LFunction(lfun_character(chi))
        sage: L(3)
        0.884023811750080

    TESTS:

    A non-primitive one::

        sage: L = LFunction(lfun_character(DirichletGroup(6).gen()))
        sage: L(4)
        0.940025680877124

    With complex arguments::

        sage: from sage.lfunctions.pari import lfun_character, LFunction
        sage: chi = DirichletGroup(6, CC).gen().primitive_character()
        sage: L = LFunction(lfun_character(chi))
        sage: L(3)
        0.884023811750080

    Check the values::

        sage: chi = DirichletGroup(24)([1, -1, -1]); chi
        Dirichlet character modulo 24 of conductor 24
        mapping 7 |--> 1, 13 |--> -1, 17 |--> -1
        sage: Lchi = lfun_character(chi)
        sage: v = [0] + Lchi.lfunan(30).sage()
        sage: all(v[i] == chi(i) for i in (7, 13, 17))
        True
    """
    if not chi.is_primitive():
        chi = chi.primitive_character()
    G, v = chi._pari_init_()
    return pari.lfuncreate([G, v])


def lfun_hgm(motif, t):
    """
    Create the `L`-function of an hypergeometric motive.

    OUTPUT: one :pari:`lfun` object

    EXAMPLES::

        sage: from sage.lfunctions.pari import lfun_hgm, LFunction
        sage: from sage.modular.hypergeometric_motive import HypergeometricData as Hyp
        sage: H = Hyp(gamma_list=([3,-1,-1,-1]))
        sage: L = LFunction(lfun_hgm(H, 1/5))
        sage: L(3)
        0.901925346034773
    """
    H = pari.hgminit(*motif.alpha_beta())
    lf = pari.lfunhgm(H, t)
    return pari.lfuncreate(lf)


def lfun_elliptic_curve(E):
    """
    Create the `L`-function of an elliptic curve.

    OUTPUT: one :pari:`lfun` object

    EXAMPLES::

        sage: from sage.lfunctions.pari import lfun_elliptic_curve, LFunction
        sage: E = EllipticCurve('11a1')
        sage: L = LFunction(lfun_elliptic_curve(E))
        sage: L(3)
        0.752723147389051
        sage: L(1)
        0.253841860855911

    Over number fields::

        sage: K.<a> = QuadraticField(2)
        sage: E = EllipticCurve([1, a])
        sage: L = LFunction(lfun_elliptic_curve(E))
        sage: L(3)
        1.00412346717019
    """
    return pari.lfuncreate(E)


def lfun_number_field(K):
    """
    Create the Dedekind zeta function of a number field.

    OUTPUT: one :pari:`lfun` object

    EXAMPLES::

        sage: from sage.lfunctions.pari import lfun_number_field, LFunction
        sage: L = LFunction(lfun_number_field(QQ))
        sage: L(3)
        1.20205690315959

        sage: K = NumberField(x**2 - 2, 'a')
        sage: L = LFunction(lfun_number_field(K))
        sage: L(3)
        1.15202784126080
        sage: L(0)
        ...0.000000000000000
    """
    return pari.lfuncreate(K)


def lfun_eta_quotient(scalings, exponents):
    """
    Return the `L`-function of an eta-quotient.

    This uses :pari:`lfunetaquo`.

    INPUT:

    - ``scalings`` -- list of integers; the scaling factors

    - ``exponents`` -- list of integers; the exponents

    EXAMPLES::

        sage: from sage.lfunctions.pari import lfun_eta_quotient, LFunction
        sage: L = LFunction(lfun_eta_quotient([1], [24]))
        sage: L(1)
        0.0374412812685155

        sage: lfun_eta_quotient([6], [4])
        [[Vecsmall([7]), [Vecsmall([6]), Vecsmall([4]), 0]], 0, [0, 1], 2, 36, 1]

        sage: lfun_eta_quotient([2, 1, 4], [5, -2, -2])
        Traceback (most recent call last):
        ...
        PariError: sorry, noncuspidal eta quotient is not yet implemented

    TESTS::

        sage: lfun_eta_quotient([2, 1, 4], [5])
        Traceback (most recent call last):
        ...
        ValueError: arguments should have the same length
    """
    from sage.matrix.constructor import matrix
    N = len(scalings)
    if N != len(exponents):
        raise ValueError('arguments should have the same length')
    m = matrix(ZZ, N, 2, list(zip(scalings, exponents)))
    return pari.lfunetaquo(m)


def lfun_delta():
    """
    Return the `L`-function of Ramanujan's Delta modular form.

    EXAMPLES::

        sage: from sage.lfunctions.pari import lfun_delta, LFunction
        sage: L = LFunction(lfun_delta())
        sage: L(1)
        0.0374412812685155
    """
    return lfun_eta_quotient([1], [24])


def lfun_eisenstein(j, algorithm="mf"):
    """
    Return the `L`-function of the Eisenstein form `E_j`.

    INPUT:

    - ``j`` -- an even integer, at least `4`

    EXAMPLES::

        sage: from sage.lfunctions.pari import lfun_eisenstein, LFunction
        sage: L = LFunction(lfun_eisenstein(16))
        sage: L(1)
        -0.291657724743874
        sage: L = LFunction(lfun_eisenstein(20))
        sage: L(2)
        -5.02355351645998

    TESTS::

        sage: from sage.lfunctions.pari import lfun_eisenstein, LFunction
        sage: L = LFunction(lfun_eisenstein(16, "shift"))
        sage: L(1)
        -0.291657724743874
    """
    # this produces the function zeta(s) zeta(s-j+1)

    if algorithm == "shift":
        L = pari.lfunshift(1, j - 1, 1)
        L[2] = [0, 1]  # this is a hack, remove once no longer needed
        return L

    Ej = pari.mfEk(j)
    Ej = pari.mflinear([Ej], [1 / pari.mfcoef(Ej, 1)])  # needed to scale
    mf = pari.mfinit(Ej)
    return pari.lfunmf(mf, Ej)


def lfun_modular_form(f):
    """
    Return the `L`-function of the modular form `f`.

    INPUT:

    - `f` -- a modular form

    EXAMPLES::

        sage: from sage.lfunctions.pari import lfun_modular_form, LFunction
        sage: h = Newforms(37)[1]
        sage: L = LFunction(lfun_modular_form(h))
        sage: L(1)
        0.725681061936153
    """
    mf = pari.mfinit(f)
    return pari.lfunmf(mf, f)


def lfun_quadratic_form(qf):
    """
    Return the `L`-function of a positive definite quadratic form.

    This uses :pari:`lfunqf`.

    EXAMPLES::

        sage: from sage.lfunctions.pari import lfun_quadratic_form, LFunction
        sage: Q = QuadraticForm(ZZ, 2, [2, 3, 4])
        sage: L = LFunction(lfun_quadratic_form(Q))
        sage: L(3)
        0.377597233183583
    """
    if not qf.is_positive_definite():
        raise ValueError('quadratic form must be positive definite')
    return pari.lfunqf(qf.matrix())


def lfun_genus2(C):
    """
    Return the `L`-function of a curve of genus 2.

    INPUT:

    - ``C`` -- hyperelliptic curve of genus 2

    Currently, the model needs to be minimal at 2.

    This uses :pari:`lfungenus2`.

    EXAMPLES::

        sage: from sage.lfunctions.pari import lfun_genus2, LFunction
        sage: x = polygen(QQ, 'x')
        sage: C = HyperellipticCurve(x^5 + x + 1)
        sage: L = LFunction(lfun_genus2(C))  # this one is broken
        ...
        sage: L(3)
        0.965946926261520

        sage: C = HyperellipticCurve(x^2 + x, x^3 + x^2 + 1)
        sage: L = LFunction(lfun_genus2(C))
        sage: L(2)
        0.364286342944359

    TESTS::

        sage: x = polygen(QQ, 'x')
        sage: H = HyperellipticCurve(x^15 + x + 1)
        sage: L = LFunction(lfun_genus2(H))
        Traceback (most recent call last):
        ...
        ValueError: curve must be hyperelliptic of genus 2
    """
    from sage.schemes.hyperelliptic_curves.hyperelliptic_g2 import HyperellipticCurve_g2 as hyp_g2
    if not isinstance(C, hyp_g2):
        raise ValueError('curve must be hyperelliptic of genus 2')
    P, Q = C.hyperelliptic_polynomials()
    return pari.lfungenus2(P) if not Q else pari.lfungenus2([P, Q])


class LFunction(SageObject):
    r"""
    Build the `L`-function from a PARI `L`-function.

    .. RUBRIC:: Rank 1 elliptic curve

    We compute with the `L`-series of a rank `1` curve. ::

        sage: E = EllipticCurve('37a')
        sage: L = E.lseries().dokchitser(algorithm='pari'); L
        PARI L-function associated to Elliptic Curve defined by
        y^2 + y = x^3 - x over Rational Field
        sage: L(1)
        0.000000000000000
        sage: L.derivative(1)
        0.305999773834052
        sage: L.derivative(1, 2)
        0.373095594536324
        sage: L.cost()
        205
        sage: L.taylor_series(1, 4)
        0.000000000000000 + 0.305999773834052*z + 0.186547797268162*z^2
        - 0.136791463097188*z^3 + O(z^4)
        sage: L.check_functional_equation()  # abs tol 4e-19
        1.08420217248550e-19

    .. RUBRIC:: Rank 2 elliptic curve

    We compute the leading coefficient and Taylor expansion of the
    `L`-series of a rank `2` elliptic curve::

        sage: E = EllipticCurve('389a')
        sage: L = E.lseries().dokchitser(algorithm='pari')
        sage: L.cost()
        666
        sage: L.derivative(1, E.rank())
        1.51863300057685
        sage: L.taylor_series(1, 4)
        ...e-19 + (...e-19)*z + 0.759316500288427*z^2 - 0.430302337583362*z^3 + O(z^4)

    .. RUBRIC:: Number field

    We compute with the Dedekind zeta function of a number field::

        sage: x = var('x')
        sage: K = NumberField(x**4 - x**2 - 1,'a')
        sage: L = K.zeta_function(algorithm='pari')
        sage: L.conductor
        400
        sage: L.cost()
        5499
        sage: L(2)
        1.10398438736918
        sage: L.taylor_series(2, 3)
        1.10398438736918 - 0.215822638498759*z + 0.279836437522536*z^2 + O(z^3)

    .. RUBRIC:: Ramanujan `\Delta` L-function

    The coefficients are given by Ramanujan's tau function::

        sage: from sage.lfunctions.pari import lfun_generic, LFunction
        sage: lf = lfun_generic(conductor=1, gammaV=[0, 1], weight=12, eps=1)
        sage: tau = pari('k->vector(k,n,(5*sigma(n,3)+7*sigma(n,5))*n/12 - 35*sum(k=1,n-1,(6*k-4*(n-k))*sigma(k,3)*sigma(n-k,5)))')
        sage: lf.init_coeffs(tau)
        sage: L = LFunction(lf)

    Now we are ready to evaluate, etc. ::

        sage: L(1)
        0.0374412812685155
        sage: L.taylor_series(1, 3)
        0.0374412812685155 + 0.0709221123619322*z + 0.0380744761270520*z^2 + O(z^3)
    """
    def __init__(self, lfun, prec=None, max_im=1) -> None:
        """
        Initialization of the `L`-function from a PARI `L`-function.

        INPUT:

        - ``lfun`` -- a PARI :pari:`lfun` object or an instance of :class:`lfun_generic`
        - ``prec`` -- integer (default: 53); number of *bits* of precision

        - ``max_im`` -- real (default: 1); maximal imaginary part

        EXAMPLES::

            sage: from sage.lfunctions.pari import lfun_generic, LFunction
            sage: lf = lfun_generic(conductor=1, gammaV=[0], weight=1, eps=1, poles=[1], residues=[-1], v=pari('k->vector(k,n,1)'))
            sage: L = LFunction(lf)
            sage: L.cost()
            8
        """
        if isinstance(lfun, lfun_generic):
            # preparation using motivic data
            self._L = lfun.__pari__()
        elif isinstance(lfun, Gen):
            # already some PARI lfun
            self._L = lfun
        else:
            # create a PARI lfunction from other input data
            self._L = pari.lfuncreate(lfun)

        self._conductor = ZZ(self._L[4])
        self._weight = ZZ(self._L[3])
        self._eps = ZZ(self._L[5])

        self._max_im = max_im

        if prec is None:
            self.prec = 53
        else:
            self.prec = PyNumber_Index(prec)
        self._RR = RealField(self.prec)
        self._CC = ComplexField(self.prec)
        # Complex field used for inputs, which ensures exact 1-to-1
        # conversion to/from PARI. Since PARI objects have a precision
        # in machine words (not bits), this is typically higher. For
        # example, the default of 53 bits of precision would become 64.
        self._CCin = ComplexField(pari.bitprecision(self._RR(1)))

    def _repr_(self) -> str:
        """
        EXAMPLES::

            sage: from sage.lfunctions.pari import lfun_number_field, LFunction
            sage: L = LFunction(lfun_number_field(QQ)); L
            L-series of conductor 1 and weight 1
        """
        return "L-series of conductor %s and weight %s" % (self._conductor,
                                                           self._weight)

    @property
    def conductor(self):
        """
        Return the conductor.

        EXAMPLES::

            sage: from sage.lfunctions.pari import *
            sage: L = LFunction(lfun_number_field(QQ)); L.conductor
            1
            sage: E = EllipticCurve('11a')
            sage: L = LFunction(lfun_number_field(E)); L.conductor
            11
        """
        return self._conductor

    def sign(self):
        """
        Return the sign.

        This complex number is the sign of the functional equation,
        sometimes also called the root number.

        See https://www.lmfdb.org/knowledge/show/lfunction.sign

        EXAMPLES::

            sage: from sage.lfunctions.pari import *
            sage: L = LFunction(lfun_number_field(QQ)); L.sign()
            1
        """
        return self._eps

    def cost(self, domain=None):
        """
        Return number of coefficients `a_n` that are needed in
        order to perform most relevant `L`-function computations to
        the desired precision.

        INPUT:

        - ``domain`` -- optional triple (center, width, height)

        The domain is then a rectangle around the real point ``center``
        with size ``2*width`` and ``2*height``.

        For computation with real arguments, one should set ``height`` to zero.

        EXAMPLES::

            sage: E = EllipticCurve('11a')
            sage: L = E.lseries().dokchitser(algorithm='pari')
            sage: L.cost()
            112
            sage: E = EllipticCurve('5077a')
            sage: L = E.lseries().dokchitser(algorithm='pari')
            sage: L.cost()
            2406

            sage: from sage.lfunctions.pari import lfun_generic, LFunction
            sage: lf = lfun_generic(conductor=1, gammaV=[0], weight=1, eps=1,
            ....:        poles=[1], residues=[-1], v=pari('k->vector(k,n,1)'))
            sage: L = LFunction(lf)
            sage: L.cost()
            8
        """
        # domain syntax :
        #    [center, width, height]
        # or [h] then [k/2, 0, h]
        # or [w, h] then [k/2, w, h]
        # where k/2 is the real point of the symmetry line

        if domain is None:
            domain = [self._weight / 2, self._max_im]

        # it could be that this precision-manipulation is useless
        num = ZZ(pari.lfuncost(self._L, domain, self.prec)[0])

        return num

    num_coeffs = cost

    def Lambda(self, s):
        """
        Evaluate the completed `L`-function at s.

        EXAMPLES::

            sage: from sage.lfunctions.pari import lfun_number_field, LFunction
            sage: L = LFunction(lfun_number_field(QQ))
            sage: L.Lambda(2)
            0.523598775598299
            sage: L.Lambda(1 - 2)
            0.523598775598299
        """
        s = self._CCin(s)
        R = self._CC
        return R(pari.lfunlambda(self._L, s, precision=self.prec))

    def hardy(self, t):
        """
        Evaluate the associated Hardy function at t.

        This only works for real t.

        EXAMPLES::

            sage: from sage.lfunctions.pari import lfun_number_field, LFunction
            sage: L = LFunction(lfun_number_field(QQ))
            sage: L.hardy(2)
            -0.962008487244041

        TESTS::

            sage: L.hardy(.4 + .3*I)
            Traceback (most recent call last):
            ...
            PariError: incorrect type in lfunhardy (t_COMPLEX)
        """
        t = self._CCin(t)
        R = self._RR
        return R(pari.lfunhardy(self._L, t, precision=self.prec))

    def derivative(self, s, D=1):
        """
        Return the derivative of the `L`-function at point s and order D.

        INPUT:

        - ``s`` -- complex number

        - ``D`` -- integer (default: 1)

        EXAMPLES::

            sage: from sage.lfunctions.pari import lfun_number_field, LFunction
            sage: L = LFunction(lfun_number_field(QQ))
            sage: L.derivative(2)
            -0.937548254315844
        """
        s = self._CCin(s)
        R = self._CC
        return R(pari.lfun(self._L, s, D, precision=self.prec))

    def taylor_series(self, s, k=6, var='z'):
        r"""
        Return the first `k` terms of the Taylor series expansion
        of the `L`-series about `s`.

        This is returned as a formal power series in ``var``.

        INPUT:

        - ``s`` -- complex number; point about which to expand

        - ``k`` -- integer (default: 6); series is `O(``var``^k)`

        - ``var`` -- string (default: ``'z'``); variable of power series

        EXAMPLES::

            sage: from sage.lfunctions.pari import lfun_number_field, LFunction
            sage: lf = lfun_number_field(QQ)
            sage: L = LFunction(lf)
            sage: L.taylor_series(2, 3)
            1.64493406684823 - 0.937548254315844*z + 0.994640117149451*z^2 + O(z^3)
            sage: E = EllipticCurve('37a')
            sage: L = E.lseries().dokchitser(algorithm='pari')
            sage: L.taylor_series(1)
            0.000000000000000 + 0.305999773834052*z + 0.186547797268162*z^2
            - 0.136791463097188*z^3 + 0.0161066468496401*z^4 + 0.0185955175398802*z^5 + O(z^6)

        We compute a Taylor series where each coefficient is to high
        precision::

            sage: E = EllipticCurve('389a')
            sage: L = E.lseries().dokchitser(200,algorithm='pari')
            sage: L.taylor_series(1, 3)
            2...e-63 + (...e-63)*z + 0.75931650028842677023019260789472201907809751649492435158581*z^2 + O(z^3)

        Check that :issue:`25402` is fixed::

            sage: L = EllipticCurve("24a1").modular_form().lseries()
            sage: L.taylor_series(-1, 3)
            0.000000000000000 - 0.702565506265199*z + 0.638929001045535*z^2 + O(z^3)
        """
        pt = pari.Ser([s, 1], d=k)  # s + x + O(x^k)
        B = PowerSeriesRing(self._CC, var)
        return B(pari.lfun(self._L, pt, precision=self.prec))

    def zeros(self, maxi):
        """
        Return the zeros with imaginary part bounded by ``maxi``.

        EXAMPLES::

            sage: from sage.lfunctions.pari import lfun_number_field, LFunction
            sage: lf = lfun_number_field(QQ)
            sage: L = LFunction(lf)
            sage: L.zeros(20)
            [14.1347251417347]
        """
        R = self._CC
        return [R(z) for z in pari.lfunzeros(self._L, maxi)]

    def _clear_value_cache(self) -> None:
        """
        Clear the cache where values of the function are stored.

        EXAMPLES::

            sage: from sage.lfunctions.pari import lfun_number_field, LFunction
            sage: lf = lfun_number_field(QQ)
            sage: L = LFunction(lf)
            sage: L(4)
            1.08232323371114
            sage: L._clear_value_cache()
        """
        del self.__values

    def __call__(self, s):
        r"""
        Return the value of the `L`-function at point ``s``.

        INPUT:

        - ``s`` -- complex number

        .. NOTE::

            Evaluation of the function takes a long time, so each
            evaluation is cached. Call :meth:`_clear_value_cache` to
            clear the evaluation cache.

        EXAMPLES::

            sage: E = EllipticCurve('5077a')
            sage: L = E.lseries().dokchitser(100, algorithm='pari')
            sage: L(1)
            0.00000000000000000000000000000
            sage: L(1 + I)
            -1.3085436607849493358323930438 + 0.81298000036784359634835412129*I
        """
        s = self._CC(s)
        try:
            return self.__values[s]
        except AttributeError:
            self.__values = {}
        except KeyError:
            pass
        try:
            value = self._CC(pari.lfun(self._L, s, precision=self.prec))
        except NameError:
            raise ArithmeticError('pole here')
        else:
            return value

    def check_functional_equation(self):
        r"""
        Verify how well numerically the functional equation is satisfied.

        If what this function returns does not look like 0 at all,
        probably the functional equation is wrong, i.e. some of the
        parameters gammaV, conductor, etc., or the coefficients are wrong.

        EXAMPLES::

            sage: from sage.lfunctions.pari import lfun_generic, LFunction
            sage: lf = lfun_generic(conductor=1, gammaV=[0], weight=1, eps=1,
            ....:          poles=[1], residues=[1], v=pari('k->vector(k,n,1)'))
            sage: L = LFunction(lf)
            sage: L.check_functional_equation()
            4.33680868994202e-19

        If we choose the sign in functional equation for the
        `\zeta` function incorrectly, the functional equation
        does not check out::

            sage: lf = lfun_generic(conductor=1, gammaV=[0], weight=1,
            ....:                   eps=-1, poles=[1], residues=[1])
            sage: lf.init_coeffs([1]*2000)
            sage: L = LFunction(lf)
            sage: L.check_functional_equation()
            16.0000000000000
        """
        if not self._max_im:
            quality = pari.lfuncheckfeq(self._L, 335/339)
        else:
            # check by pari at 335/339 + I/7
            quality = pari.lfuncheckfeq(self._L)
        return self._RR(2)**quality
